__all__ = [
    "Asinh",
    "Isymlog",
    "Log1p",
    "OnePolynomial",
    "Safeexp",
    "Safelog",
    "SequenceToBatch",
    "Sinh",
    "Squeeze",
    "Symlog",
    "ToBinaryLabel",
    "ToCategoricalLabel",
    "ToFloat",
    "ToLong",
]

import math
from typing import Union

from torch import Tensor
from torch.nn import Module

from gravitorch.utils.tensor.math_ops import isymlog, safeexp, safelog, symlog


class Asinh(Module):
    r"""Implements a ``torch.nn.Module`` to compute the inverse hyperbolic sine
    (arcsinh) of the elements."""

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.asinh()


class Sinh(Module):
    r"""Implements a ``torch.nn.Module`` to compute the hyperbolic sine (sinh)
    of the elements."""

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.sinh()


class Log1p(Module):
    r"""Implements a ``torch.nn.Module`` to compute the natural logarithm of
    ``(1 + input)``."""

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.log1p()


class OnePolynomial(Module):
    r"""Implements a ``torch.nn.Module`` to compute a polynomial transformation
    with a single term.

    The equation is ``alpha * x^gamma + beta`` where ``x`` is the
    module input tensor.

    Args:
        alpha (float, optional): Specifies the alpha value.
            Default: ``1.0``
        beta (float, optional): Specifies the beta value.
            Default: ``0.0``
        gamma (float, optional): Specifies the gamma value.
            Default: ``1.0``
    """

    def __init__(self, alpha: float = 1.0, beta: float = 0.0, gamma: float = 1.0):
        super().__init__()
        self._alpha = float(alpha)
        self._beta = float(beta)
        self._gamma = float(gamma)

    def extra_repr(self) -> str:
        return f"alpha={self._alpha}, beta={self._beta}, gamma={self._gamma}"

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Computes a polynomial transformation of the input tensor.

        Args:
            tensor (``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn)``): Specifies the tensor of values
                to transform.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn)``: The transformed values.
        """
        return tensor.pow(self._gamma).mul(self._alpha).add(self._beta)

    @classmethod
    def create_from_range(
        cls,
        gamma: float = 1.0,
        input_min_value: float = 0.0,
        input_max_value: float = 1.0,
        output_min_value: float = 0.0,
        output_max_value: float = 1.0,
    ) -> "OnePolynomial":
        r"""Instantiates a polynomial transform module for a given input and
        output ranges.

        Args:
            gamma (float, optional): Specifies the gamma value.
                Default: ``1.0``
            input_min_value (float, optional): Specifies the expected
                minimum input value. Default: ``0.0``
            input_max_value (float, optional): Specifies the expected
                maximum input value. Default: ``1.0``
            output_min_value (float, optional): Specifies the expected
                minimum output value. Default: ``0.0``
            output_max_value (float, optional): Specifies the expected
                maximum output value. Default: ``1.0``

        Returns:
            ``OnePolynomial``: An instantiated polynomial transform
                module.
        """
        alpha = (output_max_value - output_min_value) / (
            math.pow(input_max_value, gamma) - math.pow(input_min_value, gamma)
        )
        return cls(
            alpha=alpha,
            beta=output_min_value - alpha * math.pow(input_min_value, gamma),
            gamma=gamma,
        )


class Safeexp(Module):
    r"""Implements a ``torch.nn.Module`` to compute the exponential of the
    elements.

    The values that are higher than the specified minimum value are
    set to this maximum value. Using a not too large positive value
    leads to an output tensor without Inf.

    Args:
        max_value (float, optional): Specifies the maximum value.
            Default: ``20.0``
    """

    def __init__(self, max_value: float = 20.0):
        super().__init__()
        self._max_value = float(max_value)

    @property
    def max_value(self) -> float:
        r"""float: The maximum value before to compute the exponential."""
        return self._max_value

    def extra_repr(self) -> str:
        return f"max_value={self._max_value}"

    def forward(self, tensor: Tensor) -> Tensor:
        return safeexp(tensor, self._max_value)


class Safelog(Module):
    r"""Implements a ``torch.nn.Module`` to compute the logarithm natural of the
    elements.

    The values that are lower than the specified minimum value are set
    to this minimum value. Using a small positive value leads to an
    output tensor without NaN or Inf.

    Args:
        min_value (float, optional): Specifies the minimum value.
            Default: ``1e-8``
    """

    def __init__(self, min_value: float = 1e-8):
        super().__init__()
        self._min_value = float(min_value)

    @property
    def min_value(self) -> float:
        r"""float: The minimum value before to compute the exponential."""
        return self._min_value

    def extra_repr(self) -> str:
        return f"min_value={self._min_value}"

    def forward(self, tensor: Tensor) -> Tensor:
        return safelog(tensor, self._min_value)


class Squeeze(Module):
    r"""Implements a ``torch.nn.Module`` to squeeze the input tensor.

    Args:
        dim (int or ``None``, optional): Specifies the dimension to
            squeeze the input tensor. If ``None``, all the dimensions
            of the input tensor of size 1 are removed.
            Default: ``None``
    """

    def __init__(self, dim: Union[int, None] = None):
        super().__init__()
        self._dim = dim

    def extra_repr(self) -> str:
        return f"dim={self._dim}"

    def forward(self, tensor: Tensor) -> Tensor:
        if self._dim is None:
            return tensor.squeeze()
        return tensor.squeeze(self._dim)


class Symlog(Module):
    r"""Implements a ``torch.nn.Module`` to compute the symmetric logarithm
    natural of the elements."""

    def forward(self, tensor: Tensor) -> Tensor:
        return symlog(tensor)


class Isymlog(Module):
    r"""Implements a ``torch.nn.Module`` to compute the inverse symmetric
    logarithm natural of the elements."""

    def forward(self, tensor: Tensor) -> Tensor:
        return isymlog(tensor)


class ToBinaryLabel(Module):
    r"""Implements a ``torch.nn.Module`` to compute binary labels from scores by
    thresholding.

    Args:
        threshold (float, optional): Specifies the threshold value
            used to compute the binary labels.
    """

    def __init__(self, threshold: float = 0.0):
        super().__init__()
        self._threshold = float(threshold)

    @property
    def threshold(self) -> float:
        r"""float: The threshold used to compute the binary label."""
        return self._threshold

    def extra_repr(self) -> str:
        return f"threshold={self._threshold}"

    def forward(self, scores: Tensor) -> Tensor:
        r"""Computes binary labels from scores.

        Args:
            scores (``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn)``): Specifies the scores used to
                compute the binary labels.

        Returns:
            ``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``. The computed binary labels.
                The values are ``0`` and ``1``.
        """
        return (scores > self._threshold).long()


class ToFloat(Module):
    r"""Implements a ``torch.nn.Module`` to convert a tensor to a float
    tensor."""

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Converts a tensor to a float tensor.

        Args:
            tensor (``torch.Tensor`` of shape ``(d0, d1, ..., dn)``):
                Specifies the input tensor.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn)``. The converted float tensor.
        """
        return tensor.float()


class ToLong(Module):
    r"""Implements a ``torch.nn.Module`` to convert a tensor to a long
    tensor."""

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Converts a tensor to a long tensor.

        Args:
            tensor (``torch.Tensor`` of shape ``(d0, d1, ..., dn)``):
                Specifies the input tensor.

        Returns:
            ``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``. The converted long tensor.
        """
        return tensor.long()


class ToCategoricalLabel(Module):
    r"""Implements a ``torch.nn.Module`` to compute categorical labels from
    scores."""

    def forward(self, scores: Tensor) -> Tensor:
        r"""Computes categorical labels from scores.

        Args:
            scores (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_classes)`` and type float):
                Specifies the scores used to compute the categorical
                labels.

        Returns:
            ``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``. The computed categorical labels.
                The values are in ``{0, 1, ..., num_classes-1}``.
        """
        return scores.argmax(dim=-1, keepdim=False)


class SequenceToBatch(Module):
    r"""Implements a ``torch.nn.Module`` to convert a batch of sequences to a
    batch i.e. remove the sequence dimension.

    This module supports batch first and sequence first format.
    """

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Convert a batch of sequences to a batch.

        Args:
            tensor (``torch.Tensor`` of shape
                ``(batch_size, seq_len, d2, ..., dn)`` or
                ``(seq_len, batch_size, d2, ..., dn)``):
                Specifies the tensor to convert.

        Returns:
            ``torch.Tensor`` of shape
                ``(batch_size * seq_len, d2, ..., dn)``:
                The converted tensor.
        """
        return tensor.view(-1, *tensor.shape[2:])



__all__ = ["DictBatcherIterDataPipe", "TupleBatcherIterDataPipe"]

import logging
from collections.abc import Hashable, Iterator, Sequence
from typing import Union

import torch
from torch import Tensor
from torch.utils.data import IterDataPipe

from gravitorch.data.datapipes.iter.shuffling import (
    shuffle_tensor_mapping,
    shuffle_tensors,
)
from gravitorch.data.datapipes.iter.source import SourceWrapperIterDataPipe
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.mapping import get_first_value
from gravitorch.utils.seed import get_torch_generator
from gravitorch.utils.summary import concise_summary

logger = logging.getLogger(__name__)


class DictBatcherIterDataPipe(IterDataPipe[dict]):
    r"""Implements a source DataPipe to generate batch of examples from a
    dictionary of ``torch.Tensor``s.

    Args:
        data (dict): Specifies a dictionary with the data. The
            generated batches have the same structure that this input.
        batch_size (int): Specifies the batch size.
        shuffle (bool, optional): If ``True``, the examples are
            shuffled before to create the batches. Default: ``False``
        random_seed (int, optional): Specifies the random seed used to
            shuffle the data. Default: ``11918852809641073385``
    """

    def __init__(
        self,
        datapipe_or_data: Union[IterDataPipe[dict[Hashable, Tensor]], dict[Hashable, Tensor]],
        batch_size: int,
        shuffle: bool = False,
        random_seed: int = 11918852809641073385,
    ):
        self._datapipe_or_data = datapipe_or_data
        self._batch_size = int(batch_size)
        self._shuffle = bool(shuffle)
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[dict]:
        datapipe_or_data = self._datapipe_or_data
        if not isinstance(datapipe_or_data, IterDataPipe):
            datapipe_or_data = SourceWrapperIterDataPipe([datapipe_or_data])
        for batch in datapipe_or_data:
            if self._shuffle:
                batch = shuffle_tensor_mapping(batch, generator=self._generator)
            keys = batch.keys()
            for tensors in zip(*[torch.split(value, self._batch_size) for value in batch.values()]):
                yield {key: tensor for key, tensor in zip(keys, tensors)}

    def __len__(self) -> int:
        if isinstance(self._datapipe_or_data, IterDataPipe):
            raise TypeError(f"{type(self).__qualname__} instance doesn't have valid length")
        return (
            get_first_value(self._datapipe_or_data).shape[0] + self._batch_size - 1
        ) // self._batch_size

    def __str__(self) -> str:
        desc = (
            str(self._datapipe_or_data)
            if isinstance(self._datapipe_or_data, IterDataPipe)
            else concise_summary(self._datapipe_or_data)
        )
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  batch_size={self._batch_size},\n"
            f"  shuffle={self._shuffle},\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe_or_data={str_add_indent(desc)},\n)"
        )

    @property
    def random_seed(self) -> int:
        r"""int: The random seed used to initialize the pseudo random
        generator.
        """
        return self._generator.initial_seed()


class TupleBatcherIterDataPipe(IterDataPipe[tuple[Tensor, ...]]):
    r"""Implements a source DataPipe to generate batch of examples from a tuple
    of ``torch.Tensor``s.

    Args:
        tensors (``IterDataPipe`` or sequence of ``torch.Tensor`` of
            shape ``(num_examples, *)`` where ``*`` means any number
            of dimensions): Specifies source DataPipe or a sequence of
            tensors.
        batch_size (int): Specifies the batch size.
        shuffle (bool, optional): If ``True``, the examples are
            shuffled before to create the batches. The
            shuffling is done per sequence of tensors.
            Default: ``False``
        random_seed (int, optional): Specifies the random seed used to
            shuffle the data. Default: ``13382866045483866228``
    """

    def __init__(
        self,
        datapipe_or_tensors: Union[IterDataPipe[Sequence[Tensor]], Sequence[Tensor]],
        batch_size: int,
        shuffle: bool = False,
        random_seed: int = 13382866045483866228,
    ):
        self._datapipe_or_tensors = datapipe_or_tensors
        self._batch_size = int(batch_size)
        self._shuffle = bool(shuffle)
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[tuple[Tensor, ...]]:
        datapipe_or_tensors = self._datapipe_or_tensors
        if not isinstance(datapipe_or_tensors, IterDataPipe):
            datapipe_or_tensors = SourceWrapperIterDataPipe([datapipe_or_tensors])
        for batch in datapipe_or_tensors:
            if self._shuffle:
                batch = shuffle_tensors(batch, generator=self._generator)
            yield from zip(*[torch.split(tensor, self._batch_size) for tensor in batch])

    def __len__(self) -> int:
        if isinstance(self._datapipe_or_tensors, IterDataPipe):
            raise TypeError(f"{type(self).__qualname__} instance doesn't have valid length")
        return (self._datapipe_or_tensors[0].shape[0] + self._batch_size - 1) // self._batch_size

    def __str__(self) -> str:
        desc = (
            str(self._datapipe_or_tensors)
            if isinstance(self._datapipe_or_tensors, IterDataPipe)
            else concise_summary(self._datapipe_or_tensors)
        )
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  batch_size={self._batch_size},\n"
            f"  shuffle={self._shuffle},\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe_or_tensors={str_add_indent(desc)},\n)"
        )

    @property
    def random_seed(self) -> int:
        r"""int: The random seed used to initialize the
        pseudo random generator.
        """
        return self._generator.initial_seed()



__all__ = [
    "AutoDevicePlacement",
    "BaseDevicePlacement",
    "CpuDevicePlacement",
    "CudaDevicePlacement",
    "ManualDevicePlacement",
    "MpsDevicePlacement",
    "NoOpDevicePlacement",
    "setup_device_placement",
]

from gravitorch.utils.device_placement.auto import AutoDevicePlacement
from gravitorch.utils.device_placement.base import BaseDevicePlacement
from gravitorch.utils.device_placement.manual import (
    CpuDevicePlacement,
    CudaDevicePlacement,
    ManualDevicePlacement,
    MpsDevicePlacement,
)
from gravitorch.utils.device_placement.noop import NoOpDevicePlacement
from gravitorch.utils.device_placement.utils import setup_device_placement



r"""This package contains the implementation of some experiment trackers."""

__all__ = [
    "BaseBasicExpTracker",
    "BaseExpTracker",
    "EpochStep",
    "IterationStep",
    "NoOpExpTracker",
    "NotActivatedExpTrackerError",
    "Step",
    "TensorBoardExpTracker",
    "main_process_only",
    "sanitize_metrics",
    "setup_exp_tracker",
]

from gravitorch.utils.exp_trackers.base import (
    BaseBasicExpTracker,
    BaseExpTracker,
    NotActivatedExpTrackerError,
)
from gravitorch.utils.exp_trackers.noop import NoOpExpTracker
from gravitorch.utils.exp_trackers.steps import EpochStep, IterationStep, Step
from gravitorch.utils.exp_trackers.tensorboard import TensorBoardExpTracker
from gravitorch.utils.exp_trackers.utils import (
    main_process_only,
    sanitize_metrics,
    setup_exp_tracker,
)



r"""This package contains the metric base class and some implemented
metrics."""

__all__ = [
    "AbsoluteError",
    "AbsoluteRelativeError",
    "BaseEpochMetric",
    "BaseMetric",
    "BaseStateEpochMetric",
    "BinaryAccuracy",
    "BinaryConfusionMatrix",
    "CategoricalAccuracy",
    "CategoricalConfusionMatrix",
    "CategoricalCrossEntropy",
    "EmptyMetricError",
    "LogCoshError",
    "NormalizedMeanSquaredError",
    "PaddedSequenceMetric",
    "RootMeanSquaredError",
    "SequentialMetric",
    "SquaredAsinhError",
    "SquaredError",
    "SquaredLogError",
    "SquaredSymlogError",
    "SymmetricAbsoluteRelativeError",
    "TopKAccuracy",
    "TransformedPredictionTarget",
    "VanillaMetric",
    "setup_metric",
]

from gravitorch.models.metrics.base import BaseMetric, EmptyMetricError, setup_metric
from gravitorch.models.metrics.base_epoch import BaseEpochMetric, BaseStateEpochMetric
from gravitorch.models.metrics.classification import (
    BinaryAccuracy,
    BinaryConfusionMatrix,
    CategoricalAccuracy,
    CategoricalConfusionMatrix,
    CategoricalCrossEntropy,
    TopKAccuracy,
)
from gravitorch.models.metrics.regression import (
    AbsoluteError,
    AbsoluteRelativeError,
    LogCoshError,
    NormalizedMeanSquaredError,
    RootMeanSquaredError,
    SquaredAsinhError,
    SquaredError,
    SquaredLogError,
    SquaredSymlogError,
    SymmetricAbsoluteRelativeError,
)
from gravitorch.models.metrics.sequential import SequentialMetric
from gravitorch.models.metrics.transform import TransformedPredictionTarget
from gravitorch.models.metrics.vanilla import PaddedSequenceMetric, VanillaMetric



r"""This package contains the implementation of some training loops."""

__all__ = [
    "AMPTrainingLoop",
    "AccelerateTrainingLoop",
    "BaseBasicTrainingLoop",
    "BaseTrainingLoop",
    "NoOpTrainingLoop",
    "VanillaTrainingLoop",
    "setup_training_loop",
]

from gravitorch.utils.training_loops.accelerate import AccelerateTrainingLoop
from gravitorch.utils.training_loops.amp import AMPTrainingLoop
from gravitorch.utils.training_loops.base import BaseBasicTrainingLoop, BaseTrainingLoop
from gravitorch.utils.training_loops.noop import NoOpTrainingLoop
from gravitorch.utils.training_loops.utils import setup_training_loop
from gravitorch.utils.training_loops.vanilla import VanillaTrainingLoop



r"""This module defines the absolute error metric."""

__all__ = ["AbsoluteError"]

import logging
from typing import Union

from torch import Tensor

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState
from gravitorch.nn.functional import absolute_error

logger = logging.getLogger(__name__)


class AbsoluteError(BaseStateEpochMetric):
    r"""Implements the absolute error metric.

    Args:
        mode (str): Specifies the mode (e.g ``'train'`` or ``'eval'``).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'abs_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import AbsoluteError
        >>> metric = AbsoluteError('eval')
        >>> metric
        AbsoluteErrorMetric(mode=eval, name=abs_err)
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/abs_err_mean': 0.0,
         'eval/abs_err_min': 0.0,
         'eval/abs_err_max': 0.0,
         'eval/abs_err_sum': 0.0,
         'eval/abs_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/abs_err_mean': 0.16666666666666666,
         'eval/abs_err_min': 0.0,
         'eval/abs_err_max': 1.0,
         'eval/abs_err_sum': 2.0,
         'eval/abs_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/abs_err_mean': 0.5,
         'eval/abs_err_min': 0.0,
         'eval/abs_err_max': 1.0,
         'eval/abs_err_sum': 2.0,
         'eval/abs_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "abs_err",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute error metric given a mini-batch of
        examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(absolute_error(prediction, target))



__all__ = ["SequentialParameterInitializer"]

import logging
from collections.abc import Sequence
from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_add_indent, to_torch_sequence_str
from gravitorch.utils.parameter_initializers.base import (
    BaseDefaultParameterInitializer,
    BaseParameterInitializer,
)
from gravitorch.utils.parameter_initializers.utils import setup_parameter_initializer

logger = logging.getLogger(__name__)


class SequentialParameterInitializer(BaseDefaultParameterInitializer):
    r"""Implements a parameter initializer that sequentially calls parameter
    initializer.

    Args:
        parameter_initializers: Specifies the sequence of parameter
            initializers. The sequence order defines the order of the
            call.
        show_stats (bool, optional): If ``True``, the parameter
            statistics are shown at the end of the initialization.
            Default: ``True``
    """

    def __init__(
        self,
        parameter_initializers: Sequence[Union[BaseParameterInitializer, dict]],
        show_stats: bool = True,
    ):
        super().__init__(show_stats=show_stats)
        self._parameter_initializers = tuple(
            setup_parameter_initializer(param_init) for param_init in parameter_initializers
        )

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_add_indent(to_torch_sequence_str(self._parameter_initializers))}\n)"
        )

    def _initialize(self, engine: BaseEngine) -> None:
        for i, parameter_initializer in enumerate(self._parameter_initializers):
            logger.info(f"[{i}/{len(self._parameter_initializers)}] {parameter_initializer}")
            parameter_initializer.initialize(engine)



__all__ = [
    # Nvidia
    "CUDA_VISIBLE_DEVICES",
    # PyTorch
    "GROUP_RANK",
    "LOCAL_RANK",
    "LOCAL_WORLD_SIZE",
    "MASTER_ADDR",
    "MASTER_PORT",
    "RANK",
    "ROLE_RANK",
    "ROLE_WORLD_SIZE",
    "TORCHELASTIC_MAX_RESTARTS",
    "TORCHELASTIC_RESTART_COUNT",
    "TORCHELASTIC_RUN_ID",
    "TORCH_DISTRIBUTED_ENV_VARS",
    "WORLD_SIZE",
    # SLURM
    "SLURM_DISTRIBUTED_ENV_VARS",
    "SLURM_JOB_ID",
    "SLURM_JOB_NODELIST",
    "SLURM_LOCALID",
    "SLURM_NTASKS",
    "SLURM_PROCID",
]

CUDA_VISIBLE_DEVICES = "CUDA_VISIBLE_DEVICES"

# The following constants defines the name of the environment variables
# used in PyTorch so it is not possible to change them. In particular,
# these constants are used by ``torch.distributed.run``
# https://github.com/pytorch/pytorch/blob/master/torch/distributed/run.py to launch and initialize the PyTorch
# distributed backend.
LOCAL_RANK = "LOCAL_RANK"  # The local rank.
RANK = "RANK"  # The global rank.
# The rank of the worker group. A number between 0 and ``max_nnodes``.
# When running a single worker group per node, this is the rank of the node.
GROUP_RANK = "GROUP_RANK"
# The rank of the worker across all the workers that have the same role.
# The role of the worker is specified in the ``WorkerSpec``.
ROLE_RANK = "ROLE_RANK"
LOCAL_WORLD_SIZE = (
    "LOCAL_WORLD_SIZE"  # The local world size (e.g. number of workers running locally).
)
WORLD_SIZE = "WORLD_SIZE"  # The world size (total number of workers in the job).
# The total number of workers that was launched with the same role specified in ``WorkerSpec``.
ROLE_WORLD_SIZE = "ROLE_WORLD_SIZE"
# The FQDN of the host that is running worker with rank 0; used to initialize
# the Torch Distributed backend.
MASTER_ADDR = "MASTER_ADDR"
MASTER_PORT = (
    "MASTER_PORT"  # The port on the ``MASTER_ADDR`` that can be used to host the C10d TCP store.
)
TORCHELASTIC_RESTART_COUNT = (
    "TORCHELASTIC_RESTART_COUNT"  # The number of worker group restarts so far.
)
TORCHELASTIC_MAX_RESTARTS = (
    "TORCHELASTIC_MAX_RESTARTS"  # The configured maximum number of restarts.
)
TORCHELASTIC_RUN_ID = (
    "TORCHELASTIC_RUN_ID"  # Equal to the rendezvous ``run_id`` (e.g. unique job id).
)
# The tuple of environment variable names used to initialize the PyTorch distributed backend.
TORCH_DISTRIBUTED_ENV_VARS = (
    GROUP_RANK,
    LOCAL_RANK,
    LOCAL_WORLD_SIZE,
    MASTER_ADDR,
    MASTER_PORT,
    RANK,
    ROLE_RANK,
    ROLE_WORLD_SIZE,
    TORCHELASTIC_MAX_RESTARTS,
    TORCHELASTIC_RESTART_COUNT,
    TORCHELASTIC_RUN_ID,
    WORLD_SIZE,
)

# These SLURM environment variables are used to initialize the native PyTorch
# distributed backends. They are used by PyTorch Ignite
# (see https://github.com/sdesrozis/why-ignite for more information on how these
# environment variables were chosen)
SLURM_JOB_ID = "SLURM_JOB_ID"
SLURM_JOB_NODELIST = "SLURM_JOB_NODELIST"
SLURM_LOCALID = "SLURM_LOCALID"
SLURM_NTASKS = "SLURM_NTASKS"
SLURM_PROCID = "SLURM_PROCID"
# The tuple of environment variable names used to initialize the
# PyTorch distributed backend in a SLURM environment.
SLURM_DISTRIBUTED_ENV_VARS = (
    SLURM_JOB_ID,
    SLURM_JOB_NODELIST,
    SLURM_LOCALID,
    SLURM_NTASKS,
    SLURM_PROCID,
)



r"""This module implements a data summary for discrete numerical values."""

__all__ = ["BaseDiscreteDataSummary"]

from collections import Counter
from typing import Optional, TypeVar

from gravitorch.utils.data_summary.base import BaseDataSummary, EmptyDataSummaryError

T = TypeVar("T")


class BaseDiscreteDataSummary(BaseDataSummary[T]):
    r"""Implements a data summary for discrete numerical values.

    This data summary assumes that the data are discrete numerical values.
    This data summary computes the following statistics:

        - ``count``: the number of values
        - ``num_unique_values``: the number of unique values
        - ``count_{}``: the number of values per unique value

    A child class has to implement the ``add`` method.
    """

    def __init__(self):
        self._counter = Counter()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def count(self) -> int:
        r"""Gets the number of values seen by the summary.

        Returns:
            int: The number of values seen by the summary.
        """
        return len(tuple(self._counter.elements()))

    def most_common(self, n: Optional[int] = None) -> list[tuple[int, int]]:
        r"""Gets a list of the ``n`` most common elements and their counts from
        the most common to the least.

        Args:
            n (int or None, optional): Specifies the number of elements to return.
                If ``n`` is ``None``, this method returns all elements in the counter.

        Returns:
            list: The list of the ``n`` most common elements and their counts.
                Elements with equal counts are ordered in the order first encountered.
        """
        if not self.count():
            raise EmptyDataSummaryError("The summary is empty")
        return self._counter.most_common(n)

    def reset(self) -> None:
        r"""Resets the data summary."""
        self._counter.clear()

    def summary(self) -> dict:
        r"""Gets a descriptive summary of the data.

        Returns:
            dict: The data descriptive summary.

        Raises:
            ``EmptyDataSummaryError`` is the data summary is empty.
        """
        if not self.count():
            raise EmptyDataSummaryError("The summary is empty")
        summary = {
            "count": self.count(),
            "num_unique_values": len(self._counter),
        }
        for name, count in self.most_common():
            summary[f"count_{name}"] = count
        return summary



__all__ = ["TransformedPredictionTarget"]

from typing import Optional, Union

from torch import Tensor
from torch.nn import Flatten, Identity, Module

from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import BaseMetric, setup_metric
from gravitorch.nn import Asinh, Log1p, Symlog, setup_nn_module


class TransformedPredictionTarget(BaseMetric):
    r"""Implements a metric wrapper to transform the prediction and target
    inputs.

    Args:
        metric (``BaseMetric`` or dict): Specifies the metric (or its
            configuration) to wrap to transform the prediction and
            target inputs. The forward method of the metric should
            have only two inputs: prediction and target.
        prediction_transform (``torch.nn.Module`` or dict or ``None``):
            Specifies the prediction transformation or its
            configuration. If ``None``, the identity transformation
            is used. Default: ``None``
        target_transform (``torch.nn.Module`` or dict or ``None``):
            Specifies the target transformation or its configuration.
            If ``None``, the identity transformation is used.
            Default: ``None``
    """

    def __init__(
        self,
        metric: Union[BaseMetric, dict],
        prediction_transform: Union[Module, dict, None] = None,
        target_transform: Union[Module, dict, None] = None,
    ):
        super().__init__()
        self.metric = setup_metric(metric)
        self.prediction_transform = setup_nn_module(prediction_transform or Identity())
        self.target_transform = setup_nn_module(target_transform or Identity())

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the metric to the provided engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        self.metric.attach(engine)

    def forward(self, prediction: Tensor, target: Tensor) -> Optional[dict]:
        r"""Updates the metric given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of type float): Specifies
                the tensor of predictions.
            target (``torch.Tensor`` of type float and same shape as
                ``prediction``): Specifies the tensor of targets.

        Returns:
            dict or ``None``: The output of the metrics.
        """
        return self.metric(self.prediction_transform(prediction), self.target_transform(target))

    def reset(self) -> None:
        r"""Resets all the metrics."""
        self.metric.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and logs the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine``, optional): Specifies the engine.
                This argument is required to log the results in the
                engine. Default: ``None``.

        Returns:
             dict: The results of the metric
        """
        return self.metric.value(engine)

    @classmethod
    def create_asinh(cls, metric: Union[BaseMetric, dict]) -> "TransformedPredictionTarget":
        r"""Creates a ``TransformedMetric`` with inverse hyperbolic sine
        (arcsinh) transformations.

        Args:
            metric (``BaseMetric`` or dict): Specifies the metric (or
                its configuration) to wrap to transform the prediction
                and target inputs. The forward method of the metric
                should have only two inputs: prediction and target.

        Returns:
            ``TransformedPredictionTarget``: An instantiated metric with
                inverse hyperbolic sine (arcsinh) transformations.
        """
        return cls(metric=metric, prediction_transform=Asinh(), target_transform=Asinh())

    @classmethod
    def create_flatten(cls, metric: Union[BaseMetric, dict]) -> "TransformedPredictionTarget":
        r"""Creates a ``TransformedMetric`` with flatten transformations.

        Args:
            metric (``BaseMetric`` or dict): Specifies the metric (or
                its configuration) to wrap to transform the prediction
                and target inputs. The forward method of the metric
                should have only two inputs: prediction and target.

        Returns:
            ``TransformedPredictionTarget``: An instantiated metric
                with ``flatten`` transformations.
        """
        return cls(metric=metric, prediction_transform=Flatten(), target_transform=Flatten())

    @classmethod
    def create_log1p(cls, metric: Union[BaseMetric, dict]) -> "TransformedPredictionTarget":
        r"""Creates a ``TransformedMetric`` with log1p transformations.

        Args:
            metric (``BaseMetric`` or dict): Specifies the metric (or
                its configuration) to wrap to transform the prediction
                and target inputs. The forward method of the metric
                should have only two inputs: prediction and target.

        Returns:
            ``TransformedPredictionTarget``: An instantiated metric
                with ``log1p`` transformations.
        """
        return cls(metric=metric, prediction_transform=Log1p(), target_transform=Log1p())

    @classmethod
    def create_symlog(cls, metric: Union[BaseMetric, dict]) -> "TransformedPredictionTarget":
        r"""Creates a ``TransformedMetric`` with symlog transformations.

        Args:
            metric (``BaseMetric`` or dict): Specifies the metric (or
                its configuration) to wrap to transform the prediction
                and target inputs. The forward method of the metric
                should have only two inputs: prediction and target.

        Returns:
            ``TransformedPredictionTarget``: An instantiated metric
                with ``symlog`` transformations.
        """
        return cls(metric=metric, prediction_transform=Symlog(), target_transform=Symlog())



r"""This module defines the base engine state."""

__all__ = ["BaseEngineState"]

from abc import ABC, abstractmethod
from typing import Any, Optional

from objectory import AbstractFactory

from gravitorch.utils.history import BaseHistory


class BaseEngineState(ABC, metaclass=AbstractFactory):
    r"""Defines the base engine state.

    A state should implement the following attributes:

    .. code-block:: python

        >>> from gravitorch.utils.engine_states import BaseEngineState
        >>> state: BaseEngineState = ...  # Create an engine state
        >>> state.epoch             # 0-based, the first epoch is 0. -1 means the training has not started
        >>> state.iteration         # 0-based, the first iteration is 0. -1 means the training has not started
        >>> state.max_epochs        # maximum number of epochs to run
        >>> state.random_seed       # random seed
    """

    @property
    @abstractmethod
    def epoch(self) -> int:
        r"""Gets the epoch value.

        The epoch is 0-based, i.e. the first epoch is 0.
        The value ``-1`` is used to indicate the training has not
        started.

        Returns:
            int: The epoch value.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.epoch
            -1
        """

    @property
    @abstractmethod
    def iteration(self) -> int:
        r"""Gets the iteration value.

        The iteration is 0-based, i.e. the first iteration is 0.
        The value ``-1`` is used to indicate the training has not
        started.

        Returns:
            int: The iteration value.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.iteration
            -1
        """

    @property
    @abstractmethod
    def max_epochs(self) -> int:
        r"""Gets the maximum number of training epochs.

        Returns:
            int: The maximum number of training epochs.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.max_epochs
            1
        """

    @property
    @abstractmethod
    def random_seed(self) -> int:
        r"""Gets the random seed.

        Returns:
            int: The random seed.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.random_seed
            42
        """

    @abstractmethod
    def add_history(self, history: BaseHistory, key: Optional[str] = None) -> None:
        r"""Adds a history to the state.

        Args:
            history (``BaseHistory``): Specifies the history
                to add to the state.
            key (str or ``None``, optional): Specifies the key to store
                the history. If ``None``, the name of the history is
                used. Default: ``None``

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_history(MinScalarHistory("loss"))
            >>> state.add_history(MinScalarHistory("loss"), 'my key')
        """

    @abstractmethod
    def add_module(self, name: str, module: Any) -> None:
        r"""Adds a module to the engine state.

        Note that the name should be unique. If the name exists, the
        old module will be overwritten by the new module.

        Args:
            name (str): Specifies the name of the module to add to
                the engine state.
            module: Specifies the module to add to the enfine state.

        Example usage:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_module('model', nn.Linear(4, 6))
        """

    @abstractmethod
    def get_best_values(self, prefix: str = "", suffix: str = "") -> dict[str, Any]:
        r"""Gets the best value of each metric.

        This method ignores the metrics with empty history and the
        non-comparable history.

        Args:
            prefix (str): Specifies the prefix used to create the dict
                of best values. The goal of this prefix is to generate
                a name which is different from the metric name to
                avoid confusion. By default, the returned dict uses the
                same name as the metric. Default: ``''``
            suffix (str): Specifies the suffix used to create the dict
                of best values. The goal of this suffix is to generate
                a name which is different from the metric name to avoid
                confusion. By default, the returned dict uses the same
                name as the metric. Default: ``''``

        Returns:
            dict: The dict with the best value of each metric.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            # Create an engine state where the best value of 'accuracy' is 42.
            >>> state: BaseEngineState = ...
            >>> best_values = state.get_best_values()
            {'accuracy': 42}
            >>> state.get_best_values(prefix='best/')
            {'best/accuracy': 42}
            >>> state.get_best_values(suffix='/best')
            {'accuracy/best': 42}
        """

    @abstractmethod
    def get_history(self, key: str) -> BaseHistory:
        r"""Gets the history associated to a key.

        Args:
            key (str): Specifies the key of the history to retrieve.

        Returns
            ``BaseHistory``: The history if it exists,
                otherwise it returns an empty history. The created
                empty history is a ``GenericHistory``.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_history(MinScalarHistory("loss"))
            >>> state.get_history('loss')
            MinScalarHistory(name='loss', ...)
            >>> state.get_history('new_history')
            GenericHistory(name='new_history', ...)
        """

    @abstractmethod
    def get_histories(self) -> dict[str, BaseHistory]:
        r"""Gets all histories in the state.

        Returns
            ``dict``: The histories with their keys.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_history(MinScalarHistory("loss"))
            >>> state.get_histories()
            {'loss': MinScalarHistory(name='loss', ...)}
        """

    @abstractmethod
    def get_module(self, name: str) -> Any:
        r"""Gets a module.

        Args:
            name (str): Specifies the module to get.

        Returns:
            The module

        Raises:
            ValueError if the module does not exist.

        Example:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_module('model', nn.Linear(4, 6))
            >>> state.get_module('model')
            Linear(in_features=4, out_features=6, bias=True)
            >>> state.get_module('missing_module')
            ValueError
        """

    @abstractmethod
    def has_history(self, key: str) -> bool:
        r"""Indicates if the state has a history for the given key.

        Args:
            key (str): Specifies the key of the history.

        Returns
            bool: ``True`` if the history exists, ``False`` otherwise.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_history(MinScalarHistory("loss"))
            >>> state.has_history('loss')
            True
            >>> state.has_history('missing_history')
            False
        """

    @abstractmethod
    def has_module(self, name: str) -> bool:
        r"""Indicates if there is module for the given name.

        Args:
            name (str): Specifies the name to check.

        Returns:
            bool: ``True`` if the module exists, otherwise ``False``.

        Example:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_module('model', nn.Linear(4, 6))
            >>> state.has_module('model')
            True
            >>> state.has_module('missing_module')
            False
        """

    @abstractmethod
    def increment_epoch(self, increment: int = 1) -> None:
        r"""Increments the epoch value by the given value.

        Args:
            increment (int, optional): Specifies the increment for the
                epoch value. Default: ``1``

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.epoch
            0
            # Increment the epoch number by 1.
            >>> state.increment_epoch()
            >>> state.epoch
            1
            # Increment the epoch number by 10.
            >>> state.increment_epoch(10)
            >>> state.epoch
            11
        """

    @abstractmethod
    def increment_iteration(self, increment: int = 1) -> None:
        r"""Increments the iteration value by the given value.

        Args:
            increment (int, optional): Specifies the increment for the
                iteration value. Default: ``1``

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.iteration
            0
            # Increment the iteration number by 1.
            >>> state.increment_iteration()
            >>> state.iteration
            1
            # Increment the iteration number by 10.
            >>> state.increment_iteration(10)
            >>> state.iteration
            11
        """

    @abstractmethod
    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
            state_dict (dict): A dict with parameters.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            # Create a state dict. Please look at the implementation of the state_dict method
            # to know the expected structure
            >>> my_state_dict = {...}
            >>> state.load_state_dict(my_state_dict)
        """

    @abstractmethod
    def remove_module(self, name: str) -> None:
        r"""Removes a module from the state.

        Args:
            name (str): Specifies the name of the module to remove.

        Raises:
            ValueError if the module name is not found.

        Example:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.add_module('model', nn.Linear(4, 6))
            >>> state.has_module('model')
            True
            >>> state.remove_module('model')
            >>> state.has_module('model')
            False
            >>> state.remove_module('model')
            ValueError
        """

    @abstractmethod
    def state_dict(self) -> dict:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: the state values in a dict.

        Example:

        .. code-block:: python

            >>> from gravitorch.utils.engine_states import BaseEngineState
            >>> state: BaseEngineState = ...  # Create an engine state
            >>> state.state_dict()
            {...}
        """



__all__ = [
    "BaseResource",
    "DistributedContext",
    "LogCudaMemory",
    "LogSysInfo",
    "Logging",
    "PyTorchConfig",
    "PyTorchCudaBackend",
    "PyTorchCudnnBackend",
    "setup_resource",
]

from gravitorch.rsrc.base import BaseResource, setup_resource
from gravitorch.rsrc.distributed import DistributedContext
from gravitorch.rsrc.logging import Logging
from gravitorch.rsrc.pytorch import (
    PyTorchConfig,
    PyTorchCudaBackend,
    PyTorchCudnnBackend,
)
from gravitorch.rsrc.sysinfo import LogCudaMemory, LogSysInfo



r"""This module defines the base model."""

__all__ = ["BaseModel"]

from abc import abstractmethod
from typing import Any

from objectory import AbstractFactory
from torch.nn import Module


class BaseModel(Module, metaclass=AbstractFactory):
    r"""Defines the base model.

    To be compatible with the engine, the forward function of the model
    should return a dictionary. If you want to train the model, the
    output dictionary should contain the key ``'loss'`` with the loss
    value.
    """

    @abstractmethod
    def forward(self, batch: Any) -> dict:
        r"""Defines the forward function of the model that returns a dictionary
        containing the loss value.

        Args:
            batch: The input is the batch of data returned by the data
                loader.

        Returns:
            dict: dictionary containing the loss value.
        """



from __future__ import annotations

__all__ = ["BaseDataSource", "LoaderNotFoundError"]

import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from objectory import AbstractFactory

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataSource(ABC, Generic[T], metaclass=AbstractFactory):
    r"""Defines the base class to implement a data source.

    A data source object is responsible to create the data loaders.

    Note: it is an experimental class and the API may change.
    """

    @abstractmethod
    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the current data source to the provided engine.

        This method can be used to set up events or logs some stats to
        the engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> from gravitorch.engines import AlphaEngine
            >>> my_engine = AlphaEngine()  # Work with any engine
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            >>> data_source.attach(my_engine)
        """

    @abstractmethod
    def get_asset(self, asset_id: str) -> Any:
        r"""Gets a data asset from this data source.

        This method is useful to access some data variables/parameters
        that are not available before to load/preprocess the data.

        Args:
            asset_id (str): Specifies the ID of the asset.

        Returns:
            The asset.

        Raises:
            ``AssetNotFoundError`` if you try to access an asset that
                does not exist.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            >>> my_asset = data_source.get_asset('my_asset_id')
        """

    @abstractmethod
    def has_asset(self, asset_id: str) -> bool:
        r"""Indicates if the asset exists or not.

        Args:
            asset_id (str): Specifies the ID of the asset.

        Returns:
            bool: ``True`` if the asset exists, otherwise ``False``.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            >>> data_source.has_asset('my_asset_id')
            False
        """

    @abstractmethod
    def get_data_loader(self, loader_id: str, engine: BaseEngine | None = None) -> Iterable[T]:
        r"""Gets a data loader.

        Args:
            loader_id (str): Specifies the ID of the data loader to
                get.
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                data loader by using the current epoch value.
                Default: ``None``

        Returns:
            ``Iterable``: A data loader.

        Raises:
            ``LoaderNotFoundError`` if the loader does not exist.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            # Get the data loader associated to the ID 'train'
            >>> data_loader = data_source.get_data_loader('train')
            # Get a data loader that can use information from an engine
            >>> from gravitorch.engines import AlphaEngine
            >>> my_engine = AlphaEngine()  # Work with any engine
            >>> data_loader = data_source.get_data_loader('train', my_engine)
        """

    @abstractmethod
    def has_data_loader(self, loader_id: str) -> bool:
        r"""Indicates if the data source has a data loader with the given ID.

        Args:
            loader_id (str): Specifies the ID of the data loader.

        Returns:
            bool: ``True`` if the data loader exists, ``False``
                otherwise.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            # Check if the data source has a data loader for ID 'train'
            >>> data_source.has_data_loader('train')
            True or False
            # Check if the data source has a data loader for ID 'eval'
            >>> data_source.has_data_loader('eval')
            True or False
        """

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
            state_dict (dict): a dict with parameters

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            # Please take a look to the implementation of the state_dict
            # function to know the expected structure
            >>> state_dict = {...}
            >>> data_source.load_state_dict(state_dict)
        """

    def state_dict(self) -> dict:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: the state values in a dict.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import BaseDataSource
            >>> data_source: BaseDataSource = ...  # Instantiate a data source
            >>> state_dict = data_source.state_dict()
            {...}
        """
        return {}


class LoaderNotFoundError(Exception):
    r"""Raised when a loader is requested but does not exist."""



r"""This module defines dome utility functions for the experiment trackers."""

__all__ = ["setup_exp_tracker", "main_process_only", "sanitize_metrics"]

import logging
from typing import Union

from gravitorch.distributed import comm as dist
from gravitorch.utils.exp_trackers.base import BaseExpTracker
from gravitorch.utils.exp_trackers.noop import NoOpExpTracker
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_exp_tracker(exp_tracker: Union[BaseExpTracker, dict, None]) -> BaseExpTracker:
    r"""Sets up the experiment tracker.

    Args:
        exp_tracker (``BaseExpTracker`` or dict or None): Specifies
            the experiment tracker or its configuration. if ``None``,
            the ``NoOpExpTracker`` is instantiated.

    Returns:
        ``BaseExpTracker``: The experiment tracker
    """
    if exp_tracker is None:
        exp_tracker = NoOpExpTracker()
    if isinstance(exp_tracker, dict):
        logger.info(
            "Initializing an experiment tracker from its configuration... "
            f"{str_target_object(exp_tracker)}"
        )
        exp_tracker = BaseExpTracker.factory(**exp_tracker)
    return exp_tracker


def main_process_only(exp_tracker: Union[BaseExpTracker, dict, None]) -> BaseExpTracker:
    r"""Instantiates the experiment tracker only for the main process.

    The non-main processes use the no-op experiment tracker.

    Args:
        exp_tracker (``BaseExpTracker`` or dict or None): Specifies
            the experiment tracker or its configuration. if ``None``,
            the ``NoOpExpTracker`` is instantiated.

    Returns:
        ``BaseExpTracker``: The instantiated experiment tracker
    """
    if dist.is_main_process():
        return setup_exp_tracker(exp_tracker)
    return NoOpExpTracker()


def sanitize_metrics(metrics: dict) -> dict[str, Union[int, float]]:
    r"""Sanitize a dictionary of metrics.

    This function removes all the values that are not an integer or
    float.

    Args:
        metrics (dict): Specifies the metrics to sanitize.

    Returns:
        dict: The sanitized metrics.
    """
    return {str(key): value for key, value in metrics.items() if isinstance(value, (int, float))}



r"""This module implements some utility functions for the optimizers."""

from __future__ import annotations

__all__ = [
    "get_learning_rate_per_group",
    "get_weight_decay_per_group",
    "log_optimizer_parameters_per_group",
    "show_optimizer_parameters_per_group",
]

import logging
from typing import TYPE_CHECKING

from tabulate import tabulate
from torch.optim import Optimizer

from gravitorch.utils.exp_trackers import Step, sanitize_metrics
from gravitorch.utils.format import to_flat_dict

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


def get_learning_rate_per_group(optimizer: Optimizer) -> dict[int, float]:
    r"""Gets the learning rates to an optimizer.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the
            optimizer.

    Returns:
        set: The set of learning rates.
    """
    return _get_parameter_per_group(optimizer, "lr")


def get_weight_decay_per_group(optimizer: Optimizer) -> dict[int, float]:
    r"""Gets the weight decay for each group of an optimizer.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the
            optimizer.

    Returns:
        set: The set of weight decays.
    """
    return _get_parameter_per_group(optimizer, "weight_decay")


def _get_parameter_per_group(optimizer: Optimizer, key: str) -> dict:
    parameters = {}
    for i, params in enumerate(optimizer.param_groups):
        if key in params:
            parameters[i] = params[key]
    return parameters


def log_optimizer_parameters_per_group(
    optimizer: Optimizer,
    engine: BaseEngine,
    step: Step | None = None,
    prefix: str = "",
) -> None:
    r"""Logs the optimizer parameters for each group.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the
            optimizer.
        engine (``BaseEngine``): Specifies the engine used to log the
            optimizer parameters.
        step (``Step``, optional): Specifies the step used to log the
            optimizer parameters.
        prefix (``Step``, optional): Specifies the prefix used to log
            the optimizer parameters.
    """
    parameters = {}
    for i, group in enumerate(optimizer.param_groups):
        for key in sorted(group.keys()):
            if key != "params":
                parameters[f"{prefix}optimizer.group{i}.{key}"] = group[key]
    engine.log_metrics(sanitize_metrics(to_flat_dict(parameters)), step=step)


def show_optimizer_parameters_per_group(optimizer: Optimizer, tablefmt: str = "fancy_grid") -> None:
    r"""Shows the optimizer parameters for each group.

    This function uses the ``tabulate`` package to log the results in
    a table.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the
            optimizer.
        tablefmt (str, optional): Specifies the table format to show
            the optimizer information. You can find the valid formats
            at https://pypi.org/project/tabulate/.
            Default: ``'fancy_grid'``
    """
    lines = []
    for i, group in enumerate(optimizer.param_groups):
        line_group = [f"Group {i}"]
        for key in sorted(group.keys()):
            if key != "params":
                line_group.append(group[key])
        lines.append(line_group)

    headers = [key for key in sorted(optimizer.param_groups[0].keys()) if key != "params"]
    logger.info(
        f"Optimizer: parameters per group\n{tabulate(lines, headers=headers, tablefmt=tablefmt)}"
    )



r"""This module defines some path utility functions."""

__all__ = [
    "find_tar_files",
    "get_human_readable_file_size",
    "get_number_of_files",
    "get_original_cwd",
    "get_pythonpath",
    "sanitize_path",
    "working_directory",
]

import contextlib
import os
import tarfile
from pathlib import Path
from typing import Optional, Union
from urllib.parse import unquote, urlparse

import hydra
from hydra.core.hydra_config import HydraConfig

from gravitorch.utils.format import to_human_readable_byte_size


def get_original_cwd() -> Path:
    r"""Gets the original working directory the experiment was launched from.

    The problem is that Hydra change the working directory when the
    application is launched.

    Returns:
        ``pathlib.Path``: If Hydra is initialized, it returns the
            original working directory otherwise it returns the
            current working directory.
    """
    if HydraConfig.initialized():
        return sanitize_path(hydra.utils.get_original_cwd())
    return Path.cwd()


def get_pythonpath() -> Path:
    r"""Gets the value of PYTHONPATH or the original working directory if this
    value is not defined.

    Returns:
        ``pathlib.Path``: The value of the PYTHONPATH or the original
            working directory if it is not defined.
    """
    return sanitize_path(os.environ.get("PYTHONPATH", get_original_cwd()))


@contextlib.contextmanager
def working_directory(path: Path):
    r"""A context manager which changes the working directory to the given path,
    and then changes it back to its previous value on exit.

    SOURCE: https://gist.github.com/nottrobin/3d675653244f8814838a

    Args:
        path (``pathlib.Path``): Specifies the path to the temporary
            working directory.

    Example usage:

    .. code-block:: python

        # Do something in original directory
        >>> with working_directory(Path('/my/new/path')):
        ...     # Do something in new directory
        # Back to old directory
    """
    path = sanitize_path(path)
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def get_number_of_files(path: str) -> int:
    r"""Gets the number of files in a folder and its sub-folders.

    Args:
        path (str): Specifies the path to the folder.

    Returns:
        int: The number of files.
    """
    return sum([len(files) for _, _, files in os.walk(path)])


def find_tar_files(path: Path, recursive: bool = True) -> tuple[Path, ...]:
    r"""Finds the path of all the tar files in a given path.

    This function does not check if a path is a symbolic link so be
    careful if you are using a path with symbolic links.

    Args:
        path (``pathlib.Path``): Specifies the path where to look for
            the tar files.
        recursive (bool, optional): Specifies if it should also check
            the sub-folders.

    Returns:
        tuple: The tuple of path of tar files.
    """
    path = sanitize_path(path)
    if path.is_dir():
        list_files = []
        for sub_path in path.iterdir():
            is_dir = sub_path.is_dir()
            if is_dir and recursive:
                list_files.extend(find_tar_files(sub_path))
            elif not is_dir and tarfile.is_tarfile(sub_path):
                list_files.append(sub_path)
        return tuple(list_files)
    if tarfile.is_tarfile(path):
        return (path,)
    return tuple()


def sanitize_path(path: Union[Path, str]) -> Path:
    r"""Sanitizes a given path.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to
            sanitize.

    Returns:
        ``pathlib.Path``: The sanitized path.

    Example usage:

    .. code-block:: python

        # Let's assume the current path is /my/path
        >>> from pathlib import Path
        >>> from gravitorch.utils.path import sanitize_path
        >>> sanitize_path("something")
        PosixPath('/my/path/something')
        >>> sanitize_path("")
        PosixPath('/my/path')
        >>> sanitize_path(Path("something"))
        PosixPath('/my/path/something')
        >>> sanitize_path(Path("something/./../"))
        PosixPath('/my/path')
        # Support URI syntax
        >>> sanitize_path('file:///my/path/something/./../')
        PosixPath('/my/path')
    """
    if isinstance(path, str):
        # Use urlparse to parse file URI: https://stackoverflow.com/a/15048213
        path = Path(unquote(urlparse(path).path))
    return path.expanduser().resolve()


def get_human_readable_file_size(path: Union[Path, str], unit: Optional[str] = None) -> str:
    r"""Gets a human-readable representation of a file size.

    Args:
        path (``pathlib.Path`` or str): Specifies the file.
        unit (str, optional): Specifies the unit. If ``None``, the
            best unit is found automatically. The supported units
            are: ``'B'``, ``'KB'``, ``'MB'``, ``'GB'``, ``'TB'``.
            Default: ``None``

    Returns:
        str: The file size in a human-readable format.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.path import get_human_readable_file_size
        >>> get_human_readable_file_size('/my/path/data.txt')
        2.00 KB
    """
    return to_human_readable_byte_size(size=sanitize_path(path).stat().st_size, unit=unit)



__all__ = ["setup_lr_scheduler_creator"]

import logging
from typing import Union

from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.creators.lr_scheduler.vanilla import VanillaLRSchedulerCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_lr_scheduler_creator(
    creator: Union[BaseLRSchedulerCreator, dict, None]
) -> BaseLRSchedulerCreator:
    r"""Sets up the LR scheduler creator.

    The LR scheduler creator is instantiated from its configuration by
    using the ``BaseLRSchedulerCreator`` factory function.

    Args:
        creator (``BaseLRSchedulerCreator`` or dict or ``None``):
            Specifies the LR scheduler creator or its configuration.

    Returns:
        ``BaseLRSchedulerCreator``: The instantiated LR scheduler
            creator.
    """
    if creator is None:
        creator = VanillaLRSchedulerCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing the LR scheduler creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseLRSchedulerCreator.factory(**creator)
    return creator



__all__ = ["WeightedSumLoss"]

from typing import Optional, Union

from torch.nn import Module, ModuleDict

from gravitorch import constants as ct
from gravitorch.nn import setup_nn_module


class WeightedSumLoss(Module):
    r"""Implements a loss function (a.k.a. criterion) that computes the weighted
    sum of multiple loss functions.

    Args:
        criteria (``torch.nn.ModuleDict`` or dict): Specifies the
            loss functions (a.k.a. criteria)
            or their configuration.
        weights (dict or ``None``, optional): Specifies the weight
            associated to each loss function. The keys should match
            with the criteria keys. By default, the weight is ``1.0``
            if it is not specified. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.models.criterions import WeightedSumLoss
        # Example without specified weight. Weights are initialized to 1.0
        >>> criterion = WeightedSumLoss(
        ...     criteria={
        ...         "value": {'_target_': "torch.nn.MSELoss"},
        ...         "time": {'_target_': "torch.nn.L1Loss"},
        ...     },
        ... )
        >>> criterion
        WeightedSumLoss(
          weights: {'value': 1.0, 'time': 1.0}
          (criteria): ModuleDict(
            (value): MSELoss()
            (time): L1Loss()
          )
        )
        # Example with specified weight.
        >>> criterion = WeightedSumLoss(
        ...     criteria={
        ...         "value": {'_target_': "torch.nn.MSELoss"},
        ...         "time": {'_target_': "torch.nn.L1Loss"},
        ...     },
        ...     weights={"value": 2.0, "time": 0.5},
        ... )
        >>> criterion
        WeightedSumLoss(
          weights: {'value': 2.0, 'time': 0.5}
          (criteria): ModuleDict(
            (value): MSELoss()
            (time): L1Loss()
          )
        )
        # Example with partially specified weight.
        >>> criterion = WeightedSumLoss(
        ...     criteria={
        ...         "value": {'_target_': "torch.nn.MSELoss"},
        ...         "time": {'_target_': "torch.nn.L1Loss"},
        ...     },
        ...     weights={"time": 0.5},
        ... )
        >>> criterion
        WeightedSumLoss(
          weights: {'value': 1.0, 'time': 0.5}
          (criteria): ModuleDict(
            (value): MSELoss()
            (time): L1Loss()
          )
        )
    """

    def __init__(
        self,
        criteria: Union[ModuleDict, dict[str, Union[Module, dict]]],
        weights: Optional[dict[str, Union[int, float]]] = None,
    ):
        super().__init__()
        if not isinstance(criteria, ModuleDict):
            criteria = ModuleDict({key: setup_nn_module(value) for key, value in criteria.items()})
        self.criteria = criteria
        self._weights = weights or {}
        for key in self.criteria.keys():
            self._weights[key] = self._weights.get(key, 1.0)

    def extra_repr(self) -> str:
        return f"weights: {self._weights}"

    def forward(self, *args, **kwargs) -> dict:
        out = {ct.LOSS: 0.0}
        for key, criterion in self.criteria.items():
            loss = criterion(*args, **kwargs)
            if isinstance(loss, dict):
                loss = loss[ct.LOSS]
            out[f"{ct.LOSS}_{key}"] = loss * self._weights[key]
            out[ct.LOSS] += out[f"{ct.LOSS}_{key}"]
        return out



"""This module defines some classes to make PyTorch metric compatible with
``VanillaModel``."""

__all__ = ["PaddedSequenceMetric", "VanillaMetric"]

from typing import Optional, Union

import torch
from torch import Tensor

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import BaseMetric


class VanillaMetric(BaseMetric):
    r"""Implements a wrapper to make compatible most of the metrics with
    ``gravitorch.models.VanillaModel``.

    This class works for every ``BaseMetric`` instance that has two
    inputs: prediction and target.

    Args:
        metric (``BaseMetric`` or dict): Specifies the metric function
            or its configuration.
        mode (str, optional): Specifies the mode (e.g train or eval).
        prediction_key (str, optional): Specifies the key where the
            prediction is if the network output is a dict.
            Default: ``prediction``
        target_key (str, optional): Specifies the key where the target
            is in the batch input. Default: ``target``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.models.metrics import TopKAccuracy
        # Initialization with a metric object.
        >>> metric = VanillaMetric(TopKAccuracy(mode='train', topk=[1]))
        # Initialization with the config of a metric.
        >>> metric = VanillaMetric(mode='eval', metric={'_target_': 'TopKAccuracy', 'topk': [1]})
        # Customize keys.
        >>> net_out = {'next_sentence_prediction': ...}
        >>> batch = {'next_sentence_target': ...}
        >>> metric = VanillaMetric(
        ...     TopKAccuracy(mode='train', topk=[1]),
        ...     prediction_key='next_sentence_prediction',
        ...     target_key='next_sentence_target',
        ... )
    """

    def __init__(
        self,
        metric: Union[BaseMetric, dict],
        mode: Optional[str] = None,
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
    ):
        super().__init__()
        self.metric = self._setup_metric(metric=metric, mode=mode)
        self._prediction_key = prediction_key
        self._target_key = target_key

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        self.metric.attach(engine)

    def forward(self, cri_out: dict, net_out: dict, batch: dict) -> Optional[dict]:
        r"""Updates the metric given a mini-batch of examples.

        Args:
            cri_out (dict): Specifies the criterion output. (not used
                in this implementation)
            net_out (dict or ``torch.Tensor``): Specifies the network
                output which contains the prediction.
            batch (dict): Specifies the batch which contains the
                target.

        Returns:
            dict or ``None``: A dict with the metric values. ``None``
                means no metric value is returned.
        """
        return self.metric(
            self._get_prediction_from_net_out(net_out),
            self._get_target_from_batch(batch),
        )

    def reset(self) -> None:
        r"""Resets the metric."""
        self.metric.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine`` or None): Specifies the engine.

        Returns:
             dict: The results of the metric.
        """
        return self.metric.value(engine)

    def _get_prediction_from_net_out(self, net_out: dict) -> Tensor:
        r"""Gets the prediction from the network output.

        Args:
            net_out (dict): Specifies the network output which
                contains the prediction.

        Returns:
            ``torch.Tensor``: The tensor of predictions.
        """
        return net_out[self._prediction_key].detach()

    def _get_target_from_batch(self, batch: dict) -> Tensor:
        r"""Gets the target from the batch. The target is the tensor with the
        key 'target'.

        Args:
            batch (dict): Specifies the batch which contains the target.

        Returns:
            ``torch.Tensor``: The tensor of targets.
        """
        return batch[self._target_key].detach()

    def _setup_metric(self, metric: Union[BaseMetric, dict], mode: Optional[str]) -> BaseMetric:
        r"""Sets up the metric.

        If the input is a dict containing the configuration of the
        metric, it will instantiate the metric given its config by
        using the ``factory`` method of ``BaseMetric``.

        Args:
            metric (``BaseMetric`` or dict): Specifies the metric
                function or its configuration.
            mode (str, optional): Specifies the mode (e.g train or
                eval).

        Returns:
            ``BaseMetric``: an instance of the metric.
        """
        if isinstance(metric, dict):
            if mode is not None:
                metric["mode"] = mode
            metric = BaseMetric.factory(**metric)
        return metric


class PaddedSequenceMetric(VanillaMetric):
    r"""Implements a wrapper to make compatible most of the metrics with
    ``gravitorch.models.VanillaModel``.

    This metric is designed to work on sequences. This metric should
    have at least two inputs:

        - the prediction which is a ``torch.Tensor`` of shape
            ``(sequence_length, batch_size, *)`` or
            ``(batch_size, sequence_length, *)`` where ``*`` means
            any number of additional dimensions. This tensor is
            converted to a tensor of shape
            ``(sequence_length * batch_size, *)`` and then feeds to
            the wrapped metric.
        - the target which is a ``torch.Tensor`` of shape
            ``(sequence_length, batch_size, *)`` or
            ``(batch_size, sequence_length, *)`` where ``*`` means
            any number of additional dimensions.
            This tensor is converted to a tensor of shape
            ``(sequence_length * batch_size, *)`` and then feeds to
            the wrapped metric.

    The input mask is optional. If no mask is provided, all the steps
    are considered as valid. The mask is a ``torch.Tensor`` of shape
    ``(sequence_length, batch_size)`` or
    ``(batch_size, sequence_length)``. The type of the tensor can be
    ``torch.int`` or``torch.long`` or``torch.float`` or ``torch.bool``
    with the following values:

        - valid value: ``True`` or ``1`` if ``valid_value=True``,
            otherwise ``False`` or ``0``.
        - invalid value: ``False`` or ``0`` if ``valid_value=True``,
            otherwise ``True`` or ``1``.

    Args:
        metric (``BaseMetric`` or dict): Specifies the metric function
            or its configuration.
        mode (str, optional): Specifies the mode (e.g train or eval).
        prediction_key (str, optional): Specifies the key where the
            prediction is if the network output is a dict.
            Default: ``prediction``
        target_key (str, optional): Specifies the key where the target
            is in the batch input. Default: ``target``
        mask_key (str, optional): Specifies the key where the target
            is in the batch input. Default: ``target``
        valid_value (bool, optional): Indicates the valid values in
            the mask. If ``True``, the valid values are indicated by
            a ``True`` in the mask. If ``False``, the valid values are
            indicated by a ``False`` in the mask. Default: ``True``
        mask_in_batch (bool, optional): Indicates if the mask is in
            ``batch`` or ``net_out``. If ``True``, the mask is taken
            from the input ``batch``, otherwise it is taken from the
            input ``net_out``. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import PaddedSequenceMetric, AbsoluteError
        >>> metric = PaddedSequenceMetric(AbsoluteError(mode='train'))
        >>> metric(
        ...     cri_out={},
        ...     net_out={'prediction': torch.ones(2, 6, 1)},
        ...     batch={'target': torch.zeros(2, 6, 1)},
        ... )
        >>> metric.value()
        {'train/mae': 1.0, 'train/mae_num_examples': 12}
        >>> metric(
        ...     cri_out={},
        ...     net_out={'prediction': torch.ones(2, 4, 1)},
        ...     batch={
        ...         'target': torch.zeros(2, 4, 1),
        ...         'mask': torch.tensor([[True, True, False, True], [True, True, True, False]]),
        ...    }
        ... )
        >>> metric.value()
        {'train/mae': 1.0, 'train/mae_num_examples': 18}
    """

    def __init__(
        self,
        metric: Union[BaseMetric, dict],
        mode: Optional[str] = None,
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
        mask_key: str = ct.MASK,
        valid_value: bool = True,
        mask_in_batch: bool = True,
    ):
        super().__init__(
            metric=metric,
            mode=mode,
            prediction_key=prediction_key,
            target_key=target_key,
        )
        self._mask_key = mask_key
        self._valid_value = bool(valid_value)
        self._mask_in_batch = bool(mask_in_batch)

    def forward(self, cri_out: dict, net_out: dict, batch: dict) -> Optional[dict]:
        r"""Updates the metric given a mini-batch of examples.

        Args:
            cri_out (dict): Specifies the criterion output. (not used
                in this implementation)
            net_out (dict): Specifies the network output which
                contains the prediction.
            batch (dict): Specifies the batch which contains the
                target and the mask.

        Returns:
            dict or ``None``: A dict with the metric values. ``None``
                means no metric value is returned.
        """
        prediction = self._get_prediction_from_net_out(net_out)
        target = self._get_target_from_batch(batch)

        # See the batch of sequences as a batch of examples
        prediction = prediction.view(-1, *prediction.shape[2:])
        target = target.view(-1, *target.shape[2:])

        # Get the mask and remove the examples that are masked
        mask = (
            batch.get(self._mask_key, None)
            if self._mask_in_batch
            else net_out.get(self._mask_key, None)
        )
        if mask is not None:
            mask = mask.view(-1).bool()
            if not self._valid_value:
                mask = torch.logical_not(mask)
            prediction = prediction[mask]
            target = target[mask]

        return self.metric(prediction, target)



r"""This module implements a data summary for continuous numerical values."""

__all__ = ["BaseContinuousDataSummary", "prepare_quantiles"]

from collections import deque
from typing import TypeVar, Union

import torch
from torch import Tensor

from gravitorch.utils.data_summary.base import BaseDataSummary, EmptyDataSummaryError

T = TypeVar("T")

DEFAULT_QUANTILES = (0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)


class BaseContinuousDataSummary(BaseDataSummary[T]):
    r"""Implements a data summary for continuous numerical values.

    This data summary computes the following statistics:

        - ``count``: the number of values
        - ``sum``: the sum of tne values
        - ``mean``: the mean of tne values
        - ``median``: the median of tne values
        - ``std``: the standard deviation of tne values
        - ``max``: the max value
        - ``min``: the min value
        - ``quantiles``: the quantile values

    A child class has to implement the ``add`` method.

    Args:
        max_size (int, optional): Specifies the maximum size used to
            store the last values because it may not be possible to
            store all the values. This parameter is used to compute
            the median and the quantiles. Default: ``10000``
        quantiles (``torch.Tensor`` or tuple or list, optional):
            Specifies a sequence of quantiles to compute, which must
            be between 0 and 1 inclusive. Default:
            ``(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)``
    """

    def __init__(
        self,
        max_size: int = 10000,
        quantiles: Union[Tensor, tuple[float, ...], list[float]] = DEFAULT_QUANTILES,
    ):
        self._sum = 0.0
        self._count = 0.0
        self._min_value = float("inf")
        self._max_value = -float("inf")
        self._quantiles = prepare_quantiles(quantiles)
        # Store only the N last values to scale to large number of values.
        self._values = deque(maxlen=max_size)
        self.reset()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(max_size={self._values.maxlen:,}, "
            f"quantiles={self._quantiles.tolist()})"
        )

    def count(self) -> int:
        r"""Gets the number of values seen by the summary.

        Returns:
            int: The number of values seen by the summary.
        """
        return int(self._count)

    def max(self) -> float:
        r"""Gets the max value.

        Returns:
            float: The max value.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return self._max_value

    def mean(self) -> float:
        r"""Computes the mean value.

        This value is computed on all the values seen.

        Returns:
            float: The mean value.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return self._sum / self._count

    def median(self) -> float:
        r"""Computes the median value from the last values.

        If there are more values than the maximum window size, only the
        last values are used. Internally, this summary uses a deque to
        track the last values and the median value is computed on the
        values in the deque. The median is not unique for values with
        an even number of elements. In this case the lower of the two
        medians is returned.

        Returns:
            float: The median value from the last values.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return torch.tensor(list(self._values), dtype=torch.float).median().item()

    def min(self) -> float:
        r"""Gets the min value.

        Returns:
            float: The min value.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return self._min_value

    def quantiles(self) -> Tensor:
        r"""Computes the quantiles.

        If there are more values than the maximum size, only the last
        values are used. Internally, this summary uses a deque to
        track the last values and the quantiles are computed on the
        values in the deque.

        Returns:
            float: The standard deviation from the last values.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return torch.quantile(torch.tensor(tuple(self._values), dtype=torch.float), self._quantiles)

    def reset(self) -> None:
        r"""Resets the data summary."""
        self._sum = 0.0
        self._count = 0.0
        self._min_value = float("inf")
        self._max_value = -float("inf")
        self._values.clear()

    def std(self) -> float:
        r"""Computes the standard deviation.

        If there are more values than the maximum size, only the last
        values are used. Internally, this summary uses a deque to
        track the last values and the standard deviation is computed
        on the values in the deque.

        Returns:
            float: The standard deviation from the last values.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return torch.tensor(self._values, dtype=torch.float).std(dim=0).item()

    def sum(self) -> float:
        r"""Gets the sum value.

        Returns:
            float: The sum value.

        Raises:
            ``EmptyDataSummaryError`` if the summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        return self._sum

    def summary(self) -> dict:
        r"""Gets a descriptive summary of the data.

        Returns:
            dict: The data descriptive summary.

        Raises:
            ``EmptyDataSummaryError`` is the data summary is empty.
        """
        if not self._count:
            raise EmptyDataSummaryError("The summary is empty")
        summary = {
            "count": self.count(),
            "sum": self.sum(),
            "mean": self.mean(),
            "median": self.median(),
            "std": self.std(),
            "max": self.max(),
            "min": self.min(),
        }
        summary.update(
            {
                f"quantile {quantile:.3f}": value.item()
                for quantile, value in zip(self._quantiles, self.quantiles())
            }
        )
        return summary


def prepare_quantiles(quantiles: Union[Tensor, tuple[float, ...], list[float]]) -> Tensor:
    r"""Prepares the quantiles to be comaptible with ``torch.quantile``.

    Args:
        quantiles (``torch.Tensor``, tuple, list): Specifies a sequence
            of quantiles to compute, which must be between 0 and 1
            inclusive.

    Returns:
        ``torch.Tensor`` of type float and shape ``(num_quantiles,)``:
            The prepare quantiles.
    """
    if isinstance(quantiles, (list, tuple)):
        quantiles = torch.tensor(quantiles)
    return torch.sort(quantiles)[0]



r"""This module implements some random Fourier feature (RFF) layers."""

__all__ = ["SequenceGaussianRFF", "ScaleShiftSequenceGaussianRFF"]

import math

import torch
from torch import Tensor
from torch.nn import Module, Parameter

from gravitorch.nn.shift_scale import SequenceShiftScale
from gravitorch.nn.utils import get_module_device


class SequenceGaussianRFF(Module):
    r"""Implements a Gaussian random Fourier feature (RFF) encoder for
    sequences.

    This encoder was proposed in the following paper:

        Fourier Features Let Networks Learn High Frequency Functions
        in Low Dimensional Domains.
        Tancik M., Srinivasan PP., Mildenhall B., Fridovich-Keil S.,
        Raghavan N., Singhal U., Ramamoorthi R., Barron JT., Ng R.
        In NeurIPS, 2020. http://arxiv.org/pdf/2006.10739

    Args:
        input_size (int): Specifies the input size.
        output_size (int): Specifies the output size.
        sigma (float, optional): Specifies the standard deviation of
            the Gaussian distribution. Default: ``2*pi``
        trainable_params (bool, optional): If ``True``, the
            parameters sampled from the Gaussian distributed are
            optimized during training. Default: ``False``.
        batch_first (bool, optional): Indicates if the first dimension
            is the batch or the sequence. If ``True``, the input
            sequence should have the shape
            ``(batch_size, sequence_length, input_size)``, otherwise
            ``(sequence_length, batch_size, input_size)``. Note that
            the sub-modules of this layer should be compatible with
            the batch format. Default: ``False``
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        sigma: float = 2 * math.pi,
        trainable_params: bool = False,
        batch_first: bool = False,
    ):
        super().__init__()
        self._input_size = int(input_size)
        self._intermediate_output_size = output_size // 2
        self._output_size = int(output_size)
        self._sigma = float(sigma)
        self._batch_first = bool(batch_first)

        self.gaussian = Parameter(
            torch.randn(input_size, self._intermediate_output_size) * self._sigma,
            requires_grad=trainable_params,
        )

    @property
    def batch_first(self) -> bool:
        r"""bool: Indicates if the first dimension is the batch or
        the sequence.

        If ``True``, the input sequence should have the shape
        ``(batch_size, sequence_length, input_size)``, otherwise
        ``(sequence_length, batch_size, input_size)``.
        """
        return self._batch_first

    @property
    def input_size(self) -> int:
        r"""int: The input size of the sequence."""
        return self._input_size

    def extra_repr(self) -> str:
        return (
            f"input_size={self._input_size}, output_size={self._output_size}, "
            f"sigma={self._sigma}, batch_first={self._batch_first}"
        )

    def forward(self, sequence: Tensor) -> Tensor:
        r"""Computes the Fourier feature mappings of the input representation.

        Args:
            sequence (``torch.Tensor`` of type float and shape
                ``(batch_size, sequence_length, input_size)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, input_size)``
                otherwise): Specifies the
                input sequence. The values should be in the range
                ``[0, 1)``.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(sequence_length, batch_size, output_size)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, output_size)``
                otherwise: The computed Fourier feature mappings.
        """
        x_proj = torch.mm((2 * math.pi * sequence).view(-1, self._input_size), self.gaussian).view(
            *sequence.shape[:2], self._intermediate_output_size
        )
        return torch.cat([x_proj.sin(), x_proj.cos()], dim=2)

    def get_dummy_input(self, batch_size: int = 1, seq_len: int = 1) -> tuple[Tensor]:
        r"""Generates a dummy input for this network.

        Args:
            batch_size (int, optional): Specifies the batch size.
                Default: ``1``
            seq_len (int, optional): Specifies the sequence length.
                Default: ``1``

        Returns:
            A tuple of one torch tensor:
                - The sequence of continuous inputs: ``torch.Tensor``
                    of type float and shape
                    ``(sequence_length, batch_size, input_size)`` if
                    ``batch_first=False`` or
                    ``(batch_size, sequence_length, input_size)``
                    otherwise.
        """
        device = self.gaussian.data.device
        if self._batch_first:
            return (torch.rand(batch_size, seq_len, self._input_size, device=device),)
        return (torch.rand(seq_len, batch_size, self._input_size, device=device),)


class ScaleShiftSequenceGaussianRFF(Module):
    r"""Implements a Gaussian random Fourier feature (RFF) encoder for
    sequences, where the values are shifted and scaled.

    Unlike ``SequenceGaussianRFF``, this layer can deal with input
    values that are not in the range ``[0, 1)``.

    This encoder was proposed in the following paper:

        Fourier Features Let Networks Learn High Frequency Functions
        in Low Dimensional Domains.
        Tancik M., Srinivasan PP., Mildenhall B., Fridovich-Keil S.,
        Raghavan N., Singhal U., Ramamoorthi R., Barron JT., Ng R.
        In NeurIPS, 2020. http://arxiv.org/pdf/2006.10739

    Args:
        input_size (int): Specifies the input size.
        output_size (int): Specifies the output size.
        sigma (float, optional): Specifies the standard deviation of
            the Gaussian distribution. Default: ``2*pi``
        trainable_params (bool, optional): If ``True``, the parameters
            sampled from the Gaussian distributed are optimized during
            training. Default: ``False``.
        batch_first (bool, optional): Indicates if the first dimension
            is the batch or the sequence. If ``True``,
            the input sequence should have the shape
            ``(batch_size, sequence_length, input_size)``, otherwise
            ``(sequence_length, batch_size, input_size)``. Note that
            the sub-modules of this layer should be compatible with
            the batch format. Default: ``False``
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        sigma: float = 2 * math.pi,
        trainable_params: bool = False,
        batch_first: bool = False,
    ):
        super().__init__()
        self._input_size = input_size
        self._batch_first = batch_first

        self.shift_scale = SequenceShiftScale(batch_first=self._batch_first)
        self.rff = SequenceGaussianRFF(
            input_size=input_size,
            output_size=output_size,
            sigma=sigma,
            trainable_params=trainable_params,
            batch_first=batch_first,
        )

    @property
    def batch_first(self) -> bool:
        r"""bool: Indicates if the first dimension is the batch or the
        sequence. If ``True``, the input representation should have
        the shape ``(batch_size, sequence_length, input_size)``,
        otherwise ``(sequence_length, batch_size, input_size)``.
        """
        return self._batch_first

    def forward(self, sequence: Tensor, src_range: Tensor, dst_range: Tensor) -> Tensor:
        r"""Computes the Fourier feature mappings of the input representation.

        Args:
            sequence (``torch.Tensor`` of type float and shape
                ``(batch_size, sequence_length, input_size)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, input_size)``
                otherwise): Specifies the input sequence.
            src_range (``torch.Tensor`` of type float and shape
                ``(batch_size, 2, input_size)``): Specifies the min
                and max values of the source space. ``[:, 0, :]``
                indicates the min values and ``[:, 1, :]`` indicates
                the max values. For a given dimension, the min and
                max values should be different.
            dst_range (``torch.Tensor`` of type float and shape
                ``(batch_size, 2, input_size)``): Specifies the min
                and max values of the destination space. ``[:, 0, :]``
                indicates the min values and ``[:, 1, :]`` indicates
                the max values. For a given dimension, the min and
                max values should be different.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(sequence_length, batch_size, output_size)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, output_size)``
                otherwise: The computed Fourier feature mappings.
        """
        return self.rff(self.shift_scale(sequence, src_range, dst_range))

    def get_dummy_input(
        self, batch_size: int = 1, seq_len: int = 1
    ) -> tuple[Tensor, Tensor, Tensor]:
        r"""Generates a dummy input for this network.

        Args:
            batch_size (int, optional): Specifies the batch size.
                Default: ``1``
            seq_len (int, optional): Specifies the sequence length.
                Default: ``1``

        Returns:
            A tuple of three torch tensors:
                - The sequence of continuous inputs: ``torch.Tensor``
                    of type float and shape
                    ``(sequence_length, batch_size, input_size)`` if
                    ``batch_first=False`` or
                    ``(batch_size, sequence_length, input_size)``
                    otherwise.
                - The min and max values of the source space:
                    ``torch.Tensor`` of type float and shape
                    ``(batch_size, 2, input_size)``
                - The min and max values of the destination space:
                    ``torch.Tensor`` of type float and shape
                    ``(batch_size, 2, input_size)``
        """
        device = get_module_device(self.rff)
        src_range = torch.ones(batch_size, 2, self._input_size, device=device)
        src_range[:, 0] = 0
        dst_range = torch.ones(batch_size, 2, self._input_size, device=device)
        dst_range[:, 0] = 0
        if self._batch_first:
            return (
                torch.rand(batch_size, seq_len, self._input_size, device=device),
                src_range,
                dst_range,
            )
        return (
            torch.rand(seq_len, batch_size, self._input_size, device=device),
            src_range,
            dst_range,
        )



__all__ = [
    "AutoDataLoaderCreator",
    "VanillaDataLoaderCreator",
    "DistributedDataLoaderCreator",
]

from collections.abc import Callable
from typing import Optional, TypeVar, Union

import torch
from torch.utils.data import DataLoader, Dataset

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.data.dataloaders.collators.utils import setup_collator
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_add_indent

T = TypeVar("T")


class AutoDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a PyTorch data loader creator that automatically chooses the
    data loader creator based on the context.

    If the distributed package is activated, it uses the
    ``DistributedDataLoaderCreator``, otherwise it uses
    ``VanillaDataLoaderCreator``.


    Note the behavior of this class may change based on the new data
    loader creators.

    Args:
        batch_size (int, optional): Specifies the number of examples
            per batch to load. Default: ``1``
        shuffle (bool, optional): Specifies of the examples are
            shuffled or not. You should set to ``True`` to have the
            data reshuffled at every epoch. Default: ``False``
        num_workers (int, optional): Specifies the number of
            subprocesses to use for data loading. ``0`` means that
            the data will be loaded in the main process.
            Default: ``0``
        pin_memory (bool, optional): If ``True``, the data loader will
            copy Tensors into CUDA pinned memory before returning them.
            If your data elements are a custom type, or your
            :attr:`collate_fn` returns a batch that is a custom type,
            see the example below. Default: ``False``
        drop_last (bool, optional): set to ``True`` to drop the last
            incomplete batch, if the dataset size is not divisible by
            the batch size. If ``False`` and the size of dataset is
            not divisible by the batch size, then the last batch will
            be smaller. Default: ``False``
        seed (int, optional): Specifies the random seed used to
            shuffle the samples if ``shuffle=True``. Default: ``0``
        collate_fn (callable or dict or None, optional): Specifies the
            function used to merge a list of samples to form a
            mini-batch of Tensor(s). If ``None``, it uses the default
            PyTorch collate function. Default: ``None``
    """

    def __init__(
        self,
        batch_size: Optional[int] = 1,
        shuffle: bool = True,
        num_workers: int = 0,
        pin_memory: bool = False,
        drop_last: bool = False,
        seed: int = 0,
        collate_fn: Union[Callable, dict, None] = None,
    ):
        if dist.is_distributed():
            self._data_loader_creator = DistributedDataLoaderCreator(
                batch_size=batch_size,
                shuffle=shuffle,
                num_workers=num_workers,
                pin_memory=pin_memory,
                drop_last=drop_last,
                seed=seed,
                collate_fn=collate_fn,
            )
        else:
            self._data_loader_creator = VanillaDataLoaderCreator(
                batch_size=batch_size,
                shuffle=shuffle,
                num_workers=num_workers,
                pin_memory=pin_memory,
                drop_last=drop_last,
                seed=seed,
                collate_fn=collate_fn,
            )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  data_loader_creator={str_add_indent(str(self._data_loader_creator))},\n"
            ")"
        )

    def create(self, dataset: Dataset, engine: Optional[BaseEngine] = None) -> DataLoader[T]:
        return self._data_loader_creator.create(dataset=dataset, engine=engine)


class VanillaDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a simple PyTorch data loader creator.

    Note that this data loader creator uses the default samplers.
    If you need a different sampler, you will need to implement your
    own data loader creator.

    Args:
        batch_size (int, optional): Specifies the number of examples
            per batch to load. Default: ``1``
        shuffle (bool, optional): Specifies of the examples are
            shuffled or not. You should set to ``True`` to have the
            data reshuffled at every epoch. Default: ``False``
        num_workers (int, optional): Specifies the number of
            subprocesses to use for data loading. ``0`` means that
            the data will be loaded in the main process.
            Default: ``0``
        pin_memory (bool, optional): If ``True``, the data loader will
            copy Tensors into CUDA pinned memory before returning them.
            If your data elements are a custom type, or your
            :attr:`collate_fn` returns a batch that is a custom type,
            see the example below. Default: ``False``
        drop_last (bool, optional): set to ``True`` to drop the last
            incomplete batch, if the dataset size is not divisible by
            the batch size. If ``False`` and the size of dataset is
            not divisible by the batch size, then the last batch will
            be smaller. Default: ``False``
        seed (int, optional): Specifies the random seed used to
            shuffle the samples if ``shuffle=True``. Default: ``0``
        collate_fn (callable or dict or None, optional): Specifies the
            function used to merge a list of samples to form a
            mini-batch of Tensor(s). If ``None``, it uses the default
            PyTorch collate function. Default: ``None``
    """

    def __init__(
        self,
        batch_size: Optional[int] = 1,
        shuffle: bool = True,
        num_workers: int = 0,
        pin_memory: bool = False,
        drop_last: bool = False,
        seed: int = 0,
        collate_fn: Union[Callable, dict, None] = None,
    ):
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._num_workers = num_workers
        self._pin_memory = pin_memory
        self._drop_last = drop_last
        self._seed = seed

        self._collate_fn = setup_collator(collate_fn)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  batch_size={self._batch_size},\n"
            f"  shuffle={self._shuffle},\n"
            f"  num_workers={self._num_workers:,},\n"
            f"  pin_memory={self._pin_memory},\n"
            f"  drop_last={self._drop_last},\n"
            f"  seed={self._seed},\n"
            f"  collate_fn={str_add_indent(str(self._collate_fn))},\n"
            ")"
        )

    def create(self, dataset: Dataset, engine: Optional[BaseEngine] = None) -> DataLoader[T]:
        generator = torch.Generator()
        epoch = 0 if engine is None else engine.epoch
        generator.manual_seed(self._seed + epoch)
        return DataLoader(
            dataset=dataset,
            batch_size=self._batch_size,
            shuffle=self._shuffle,
            num_workers=self._num_workers,
            pin_memory=self._pin_memory,
            drop_last=self._drop_last,
            collate_fn=self._collate_fn,
            generator=generator,
        )


class DistributedDataLoaderCreator(VanillaDataLoaderCreator[T]):
    r"""Defines a simple distributed PyTorch data loader creator.

    This data loader creator uses the ``gravitorch.distributed`` package
    to distribute the example per process. Note that this data loader
    creator uses the default samplers. If you need a different sampler,
    you will need to implement your own data loader creator.
    """

    def create(self, dataset: Dataset, engine: Optional[BaseEngine] = None) -> DataLoader[T]:
        sampler = torch.utils.data.distributed.DistributedSampler(
            dataset,
            shuffle=self._shuffle,
            drop_last=self._drop_last,
            seed=self._seed,
            rank=dist.get_rank(),
            num_replicas=dist.get_world_size(),
        )
        if engine is not None:
            # In distributed mode, calling the set_epoch() method at the beginning
            # of each epoch before creating the DataLoader iterator is necessary to
            # make shuffling work properly across multiple epochs.
            # Otherwise, the same ordering will always be used.
            sampler.set_epoch(engine.epoch)
        # Sampler option is mutually exclusive with shuffle.
        return DataLoader(
            dataset=dataset,
            sampler=sampler,
            batch_size=self._batch_size,
            num_workers=self._num_workers,
            pin_memory=self._pin_memory,
            collate_fn=self._collate_fn,
        )



r"""This module defines a image folder dataset."""

__all__ = ["ImageFolderDataset"]

from unittest.mock import Mock

from gravitorch import constants as ct
from gravitorch.utils.integrations import (
    check_pillow,
    check_torchvision,
    is_torchvision_available,
)

if is_torchvision_available():
    from torchvision.datasets import ImageFolder
else:
    ImageFolder = Mock  # pragma: no cover


class ImageFolderDataset(ImageFolder):
    r"""Implements a dataset that returns a dict instead of a tuple.

    This dataset extends the ``torchvision.datasets.ImageFolder`` class.
    """

    def __init__(self, *args, **kwargs):
        check_torchvision()
        check_pillow()
        super().__init__(*args, **kwargs)

    def __getitem__(self, index: int) -> dict:
        r"""Get the image and the target of the index-th example.

        Args:
            index (int): Specifies the index of the example.

        Returns:
            dict: A dictionary with the image and the target of the
                ``index``-th example.
        """
        img, target = super().__getitem__(index)
        return {ct.INPUT: img, ct.TARGET: target}



__all__ = ["BaseDataCreator", "setup_data_creator"]

import logging
from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar, Union

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataCreator(ABC, Generic[T], metaclass=AbstractFactory):
    r"""Defines the base class to implement a data creator."""

    @abstractmethod
    def create(self, engine: Optional[BaseEngine] = None) -> T:
        r"""Creates data.

        Args:
            engine (``BaseEngine`` or ``None``): Specifies an engine.
                Default: ``None``

        Return:
            The created data.
        """


def setup_data_creator(data_creator: Union[BaseDataCreator, dict]) -> BaseDataCreator:
    r"""Sets up a data creator.

    The data creator is instantiated from its configuration by using
    the ``BaseDataCreator`` factory function.

    Args:
        data_creator (``BaseDataCreator`` or dict): Specifies the data
            creator or its configuration.

    Returns:
        ``BaseDataCreator``: The instantiated data creator.
    """
    if isinstance(data_creator, dict):
        logger.info(
            "Initializing a data creator from its configuration... "
            f"{str_target_object(data_creator)}"
        )
        data_creator = BaseDataCreator.factory(**data_creator)
    return data_creator



"""This module defines a simple model which is composed of 3 modules:

network, criterion and metrics.
"""

__all__ = ["VanillaModel"]

import logging
from pathlib import Path
from typing import Any, Union

import torch
from torch import Tensor
from torch.nn import Module, ModuleDict
from torch.nn.utils.rnn import PackedSequence

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.models.base import BaseModel
from gravitorch.models.metrics import BaseMetric
from gravitorch.models.utils import attach_module_to_engine
from gravitorch.nn.utils.factory import setup_nn_module
from gravitorch.nn.utils.state_dict import load_checkpoint_to_module
from gravitorch.utils.seed import manual_seed

logger = logging.getLogger(__name__)


class VanillaModel(BaseModel):
    r"""Implements a simple model which is composed of 3 modules: network,
    criterion and metrics.

    The assumptions of the ``VanillaModel`` are:

        - the input of the model is a dict containing the data of the
            mini-batch.
        - the network follow the ``BaseNetwork`` API.
        - the input of the criterion is the dicts containing the
            network output and the mini-batch data.
        - the criterion should return a dictionary containing the
            loss.
        - the metrics are optional, or it is possible to define a
            metric only for train or eval.
        - the input of the metric is the dicts containing the
            criterion output, network output and the mini-batch data.
        - the keys returned by the network, criterion and metrics must
            be different.

    Note that the metric names are ``'train_metric'`` and
    ``'eval_metric'`` because it is not possible to use
    ``'train'`` and ``'eval'`` because they are already used by
    PyTorch.

    Args:
        network (``torch.nn.Module`` or dict): Specifies the network
            module or its configuration.
        criterion (``torch.nn.Module`` or dict): Specifies the
            criterion module or its configuration.
        metrics (``torch.nn.ModuleDict`` or dict): Specifies the
            metrics or their configuration.
        checkpoint_path (``pathlib.Path`` or str or ``None``):
            Specifies a path to a model checkpoint. This weights in
            the checkpoint are used to initialize the model. ``None``
             means there is no checkpoint to load. Default: ``None``.
        random_seed (int, optional): Specifies the random seed.
            Default: ``6671429959452193306``
    """

    def __init__(
        self,
        network: Union[Module, dict[str, Any]],
        criterion: Union[Module, dict[str, Any]],
        metrics: Union[ModuleDict, dict[str, Any], None] = None,
        checkpoint_path: Union[Path, str, None] = None,
        random_seed: int = 6671429959452193306,
    ):
        super().__init__()
        manual_seed(random_seed)  # Fix the random seed for reproducibility purpose
        self.network = setup_nn_module(network)
        logger.info(f"network:\n{self.network}")
        self.criterion = setup_nn_module(criterion)
        logger.info(f"criterion:\n{self.criterion}")
        self.metrics = self._setup_metrics(metrics)
        logger.info(f"metrics:\n{self.metrics}")

        if checkpoint_path:
            load_checkpoint_to_module(checkpoint_path, self)

    def forward(self, batch: dict) -> dict:
        """Prepares the batch and feed it to the network, criterion and metric.

        Args:
            batch (dict): Specifies the dict containing data: inputs,
                targets, etc.

        Returns:
            out (dict): a dictionary of outputs including the loss value
        """
        net_out = self._parse_net_out(
            self.network(*tuple(batch[key] for key in self.network.get_input_names()))
        )
        cri_out = self.criterion(net_out, batch)
        met_out = self._get_metric_out(cri_out, net_out, batch)
        return self._get_model_out(net_out, cri_out, met_out)

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current model to the provided engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        attach_module_to_engine(self.network, engine)
        attach_module_to_engine(self.criterion, engine)
        for metric in self.metrics.values():
            attach_module_to_engine(metric, engine)

    def _setup_metrics(self, metrics: Union[ModuleDict, dict[str, Any], None]) -> ModuleDict:
        r"""Sets up the metrics.

        Args:
            metrics (``torch.nn.ModuleDict`` or dict): Specifies the
                metrics or their configuration.

        Returns:
            ``torch.nn.ModuleDict``: The metrics
        """
        if isinstance(metrics, ModuleDict):
            return metrics

        metrics = metrics or {}
        for key, metric in metrics.items():
            if isinstance(metric, dict):
                metric = BaseMetric.factory(**metric)
            metrics[key] = metric
        return ModuleDict(metrics)

    def _parse_net_out(self, net_out: Union[Tensor, PackedSequence, list, tuple]) -> dict:
        r"""Parses the network outputs to a dict.

        Args:
            net_out: Specifies the network output.

        Returns:
            dict: The dictionary containing the network output.
        """
        if torch.is_tensor(net_out) or isinstance(net_out, PackedSequence):
            net_out = (net_out,)
        return {
            output_key: net_out[i] for i, output_key in enumerate(self.network.get_output_names())
        }

    def _get_metric_out(self, cri_out: dict, net_out: dict, batch: dict) -> dict:
        r"""Computes the metric output.

        Args:
            cri_out (dict): Specifies the dict with the criterion
                output.
            net_out (dict): Specifies the dict with the network
                output.
            batch (dict): Specifies the dict containing data: inputs,
                targets, etc.

        Returns:
            dict: The dictionary containing the metric values. The
                dictionary can be empty.
        """
        # It is not possible to use 'train' or 'eval' because these
        # names are already used in ``torch.nn.Module``
        mode = f"{ct.TRAIN if self.training else ct.EVAL}_metric"
        met_out = {}
        if mode in self.metrics:
            met_out = self.metrics[mode](cri_out, net_out, batch)
        return met_out or {}

    def _get_model_out(self, net_out: dict, cri_out: dict, met_out: dict) -> dict:
        r"""Computes the model output by combining the outputs of the network,
        criterion and metric.

        Note that the keys have to be unique otherwise only one will
        be returned by the model.

        Args:
            net_out (dict): Specifies the dict with the network
                output.
            cri_out (dict): Specifies the dict with the criterion
                output.
            met_out (dict): Specifies the dict containing data:
                inputs, targets, etc.

        Returns:
            dict: The dictionary containing the model outputs.
        """
        out = {}
        out.update(net_out)
        out.update(cri_out)
        out.update(met_out)
        return out



r"""This module implements some comparable histories."""

__all__ = ["ComparableHistory", "MaxScalarHistory", "MinScalarHistory"]

from collections.abc import Iterable
from numbers import Number
from typing import Any, Optional, TypeVar

from gravitorch.utils.history.base import EmptyHistoryError
from gravitorch.utils.history.comparator import (
    BaseComparator,
    MaxScalarComparator,
    MinScalarComparator,
)
from gravitorch.utils.history.generic import GenericHistory

T = TypeVar("T")


class ComparableHistory(GenericHistory[T]):
    r"""Implements a comparable history.

    Args:
        name (str): Specifies the name of the history.
        comparator (BaseComparator): Specifies the comparator to use
            to find the best value.
        elements (iterable, optional): Specifies the initial elements.
            Each element is a tuple with the step and its associated
            value. Default: ``tuple()``
        max_size (int, optional): Specifies the maximum size
            of the history. Default: ``10``
        best_value (``T`` or ``None``, optional): Specifies the
            initial best value. If ``None``, the initial best value of
            the ``comparator`` is used. Default: ``None``
        improved (bool, optional): Indicates if the last value is the
            best value or not. Default: ``False``
    """

    def __init__(
        self,
        name: str,
        comparator: BaseComparator[T],
        elements: Iterable[tuple[Optional[int], T]] = tuple(),
        max_size: int = 10,
        best_value: Optional[T] = None,
        improved: bool = False,
    ):
        super().__init__(name=name, elements=elements, max_size=max_size)
        self._comparator = comparator
        self._best_value = best_value or self._comparator.get_initial_best_value()
        self._improved = bool(improved)

    # TODO: add to string

    def add_value(self, value: T, step: Optional[int] = None) -> None:
        self._improved = self.is_better(new_value=value, old_value=self._best_value)
        if self._improved:
            self._best_value = value
        super().add_value(value, step)

    def is_better(self, old_value: T, new_value: T) -> bool:
        r"""Indicates if the new value is better than the old value.

        Args:
            old_value: Specifies the old value to compare.
            new_value: Specifies the new value to compare.

        Returns:
            bool: ``True`` if the new value is better than the old
                value, otherwise ``False``.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import ComparableHistory
            >>> from gravitorch.utils.history.comparator import MaxScalarComparator
            >>> history = ComparableHistory('accuracy', comparator=MaxScalarComparator())
            >>> history.is_better(new_value=1, old_value=0)
            True
            >>> history.is_better(new_value=0, old_value=1)
            False
        """
        return self._comparator.is_better(new_value=new_value, old_value=old_value)

    def _get_best_value(self) -> T:
        if self.is_empty():
            raise EmptyHistoryError(
                "The history is empty so it is not possible to get the best value."
            )
        return self._best_value

    def _has_improved(self) -> bool:
        if self.is_empty():
            raise EmptyHistoryError("The history is empty.")
        return self._improved

    def is_comparable(self) -> bool:
        return True

    def config_dict(self) -> dict[str, Any]:
        config = super().config_dict()
        config["max_size"] = self.max_size
        config["comparator"] = self._comparator
        return config

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        super().load_state_dict(state_dict)
        self._improved = state_dict["improved"]
        self._best_value = state_dict["best_value"]

    def state_dict(self) -> dict[str, Any]:
        state = super().state_dict()
        state.update({"improved": self._improved, "best_value": self._best_value})
        return state


class MaxScalarHistory(ComparableHistory[Number]):
    r"""A specific implementation to track the max value of a scalar history.

    This history uses the ``MaxScalarComparator`` to find the
    best value of the history.

    Args:
        name (str): Specifies the name of the history.
        elements (iterable, optional): Specifies the initial elements.
            Each element is a tuple with the step and its associated
            value. Default: ``tuple()``
        max_size (int, optional): Specifies the maximum size
            of the history. Default: ``10``
        best_value (float or int or ``None``, optional): Specifies the
            initial best value. If ``None``, the initial best value of
            the ``comparator`` is used. Default: ``None``
        improved (bool, optional): Indicates if the last value is the
            best value or not. Default: ``False``
    """

    def __init__(
        self,
        name: str,
        elements: Iterable[tuple[Optional[int], T]] = tuple(),
        max_size: int = 10,
        best_value: Optional[T] = None,
        improved: bool = False,
    ):
        super().__init__(
            name=name,
            comparator=MaxScalarComparator(),
            elements=elements,
            max_size=max_size,
            best_value=best_value,
            improved=improved,
        )

    def config_dict(self) -> dict[str, Any]:
        config = super().config_dict()
        del config["comparator"]
        return config


class MinScalarHistory(ComparableHistory[Number]):
    r"""A specific implementation to track the min value of a scalar history.

    This history uses the ``MinScalarComparator`` to find the
    best value of the history.

    Args:
        name (str): Specifies the name of the history.
        elements (iterable, optional): Specifies the initial elements.
            Each element is a tuple with the step and its associated
            value. Default: ``tuple()``
        max_size (int, optional): Specifies the maximum size
            of the history. Default: ``10``
        best_value (float or int or ``None``, optional): Specifies the
            initial best value. If ``None``, the initial best value of
            the ``comparator`` is used. Default: ``None``
        improved (bool, optional): Indicates if the last value is the
            best value or not. Default: ``False``
    """

    def __init__(
        self,
        name: str,
        elements: Iterable[tuple[Optional[int], T]] = tuple(),
        max_size: int = 10,
        best_value: Optional[T] = None,
        improved: bool = False,
    ):
        super().__init__(
            name=name,
            comparator=MinScalarComparator(),
            elements=elements,
            max_size=max_size,
            best_value=best_value,
            improved=improved,
        )

    def config_dict(self) -> dict[str, Any]:
        config = super().config_dict()
        del config["comparator"]
        return config



__all__ = ["BaseDevicePlacement"]

from abc import ABC, abstractmethod
from typing import Any

from objectory import AbstractFactory


class BaseDevicePlacement(ABC, metaclass=AbstractFactory):
    r"""Defines a base class to send object on a target device.

    Note that the object should be sent to a single target device. This
    class cannot be used if you want to send an object to multiple
    target devices.
    """

    @abstractmethod
    def send(self, obj: Any) -> Any:
        r"""Sends the object on a target device.

        Args:
            obj: The object to send to the target device.

        Returns:
            The object on the target device.
        """



__all__ = ["TrainingRunner"]

import logging
from collections.abc import Sequence
from typing import Optional, Union

from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.handlers import setup_and_attach_handlers
from gravitorch.handlers.base import BaseHandler
from gravitorch.rsrc.base import BaseResource
from gravitorch.runners.resource import BaseResourceRunner
from gravitorch.utils.exp_trackers import setup_exp_tracker
from gravitorch.utils.exp_trackers.base import BaseExpTracker
from gravitorch.utils.format import (
    str_add_indent,
    to_pretty_json_str,
    to_torch_sequence_str,
)
from gravitorch.utils.seed import manual_seed

logger = logging.getLogger(__name__)


class TrainingRunner(BaseResourceRunner):
    r"""Implements a runner to train a ML model.

    Internally, this runner does the following steps:

        - set the experiment tracker
        - set the random seed
        - instantiate the engine
        - set up and attach the handlers
        - train the model with the engine

    Args:
        engine (``BaseEngine`` or dict): Specifies the engine or its
            configuration.
        handlers (list or tuple or ``None``): Specifies the list of
            handlers or their configuration. The handlers will be
            attached to the engine. If ``None``, no handler is
            attached to the engine. Default: ``None``
        exp_tracker (``BaseExpTracker`` or dict or None): Specifies
            the experiment tracker or its configuration. If ``None``,
            the no-operation experiment tracker is used.
        random_seed (int, optional): Specifies the random seed.
            Default: ``10139531598155730726``
        resources (sequence or ``None``, optional): Specifies a
            sequence of resources or their configurations.
            Default: ``None``
    """

    def __init__(
        self,
        engine: Union[BaseEngine, dict],
        handlers: Optional[Sequence[Union[BaseHandler, dict]]] = None,
        exp_tracker: Union[BaseExpTracker, dict, None] = None,
        random_seed: int = 10139531598155730726,
        resources: Optional[Sequence[Union[BaseResource, dict]]] = None,
    ):
        super().__init__(resources=resources)
        self._engine = engine
        self._handlers = tuple() if handlers is None else tuple(handlers)
        self._exp_tracker = exp_tracker
        self._random_seed = random_seed

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  random_seed={self._random_seed},\n"
            f"  engine={str_add_indent(to_pretty_json_str(self._engine))},\n"
            f"  exp_tracker={str_add_indent(to_pretty_json_str(self._exp_tracker))},\n"
            f"  handlers:\n  {str_add_indent(to_torch_sequence_str(self._handlers))},\n"
            f"  resources:\n  {str_add_indent(to_torch_sequence_str(self._resources))},\n"
            ")"
        )

    def _run(self) -> BaseEngine:
        return _run_training_pipeline(
            engine=self._engine,
            handlers=self._handlers,
            exp_tracker=self._exp_tracker,
            random_seed=self._random_seed,
        )


def _run_training_pipeline(
    engine: Union[BaseEngine, dict],
    handlers: Union[tuple[Union[BaseHandler, dict], ...], list[Union[BaseHandler, dict]]],
    exp_tracker: Union[BaseExpTracker, dict, None],
    random_seed: int = 4398892194000378040,
) -> BaseEngine:
    r"""Implements the training pipeline.

    Internally, this function does the following steps:

        - set the experiment tracker
        - set the random seed
        - instantiate the engine
        - set up and attach the handlers
        - train the model with the engine

    Args:
        engine (``BaseEngine`` or dict): Specifies the engine or its
            configuration.
        handlers (list or tuple): Specifies the list of handlers or
            their configuration.
        exp_tracker (``BaseExpTracker`` or dict or None): Specifies
            the experiment tracker or its configuration. If ``None``,
            the no-operation experiment tracker is used.
        random_seed (int, optional): Specifies the random seed.
            Default: ``4398892194000378040``

    Returns:
        ``BaseEngine``: The trained engine.
    """
    with setup_exp_tracker(exp_tracker) as tracker:
        random_seed = random_seed + dist.get_rank()
        logger.info(f"Set the random seed to {random_seed}")
        manual_seed(random_seed)

        if isinstance(engine, dict):
            tracker.log_hyper_parameters(engine)
            logger.info("Initializing the engine from its configuration...")
            engine = BaseEngine.factory(exp_tracker=tracker, **engine)

        logger.info("Adding the handlers to the engine...")
        setup_and_attach_handlers(engine, handlers)

        logger.info(f"engine:\n{engine}")
        engine.train()

    return engine



r"""This module implements an event manager."""

__all__ = ["EventManager"]

import logging
from collections import defaultdict
from typing import Optional

from gravitorch.utils.events.event_handlers import BaseEventHandler
from gravitorch.utils.format import str_add_indent, to_torch_sequence_str

logger = logging.getLogger(__name__)


class EventManager:
    r"""Implements an event manager.

    This event manager allows adding event handlers and firing events.
    An event is represented by a case-sensitive string.
    """

    def __init__(self):
        # This variable is used to store the handlers associated to each event.
        self._event_handlers = defaultdict(list)
        # This variable is used to track the last fired event name
        self._last_fired_event = None
        self.reset()

    def __repr__(self) -> str:
        if self._event_handlers:
            return (
                f"{self.__class__.__qualname__}(\n"
                f"  {str_add_indent(to_event_handlers_str(self._event_handlers))}\n"
                f"  last_fired_event={self._last_fired_event}\n"
                ")"
            )
        return (
            f"{self.__class__.__qualname__}(event_handlers=(), "
            f"last_fired_event={self._last_fired_event})"
        )

    @property
    def last_fired_event(self) -> Optional[str]:
        r"""Gets the last event name that was fired.

        Returns:
            str or ``None``: The last event name that was fired or
                ``None`` if no event was fired.
        """
        return self._last_fired_event

    def add_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Adds an event handler to an event.

        The event handler will be called everytime the event happens.

        Args:
            event (str): Specifies the event to attach the event
                handler.
            event_handler (``BaseEventHandler``): Specifies the
                event handler to attach to the event.

        Example usage:

        .. code-block:: python

            # Create an event manager
            >>> from gravitorch.utils.events import EventManager
            >>> event_manager = EventManager()
            # Add an event handler to the event manager
            >>> def hello_handler():
            ...     print('Hello!')
            ...
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> event_manager.add_event_handler('my_event', VanillaEventHandler(hello_handler))
        """
        self._event_handlers[str(event)].append(event_handler)
        logger.debug(f"Added {event_handler} to event {event}")

    def fire_event(self, event: str) -> None:
        r"""Fires the handler(s) for the given event.

        Args:
            event (str): Specifies the event to fire.

        Example usage:

        .. code-block:: python

            # Create an event manager
            >>> from gravitorch.utils.events import EventManager
            >>> event_manager = EventManager()
            # Fire the 'my_event' event
            >>> event_manager.fire_event('my_event')  # should do nothing because there is no event handler
            # Add an event handler
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> event_manager.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            # Fire the 'my_event' event
            >>> event_manager.fire_event('my_event')
            Hello!
        """
        logger.debug(f"Firing {event} event")
        self._last_fired_event = event
        for event_handler in self._event_handlers[event]:
            event_handler.handle()

    def has_event_handler(
        self, event_handler: BaseEventHandler, event: Optional[str] = None
    ) -> bool:
        r"""Indicates if a handler is registered in the event manager.

        Note that this method relies on the ``__eq__`` method of the
        input event handler to compare event handlers.

        Args:
            event_handler (``BaseEventHandler``): Specifies the eventn
                handler to check.
            event (str or ``None``): Specifies an event to check. If
                the value is ``None``, it will check all the events.
                Default: ``None``

        Example usage:

        .. code-block:: python

            # Create an event manager
            >>> from gravitorch.utils.events import EventManager
            >>> event_manager = EventManager()
            # Define a handler
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            # Check if `hello_handler` is registered in the event manager
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler))
            False
            # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            False
            # Add an event handler
            >>> event_manager.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            # Check if `hello_handler` is registered in the event manager
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler))
            True
            # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            True
            # Check if `hello_handler` is registered in the event manager for 'my_other_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_other_event')
            False
        """
        events = [event] if event else self._event_handlers
        for evnt in events:
            for evnt_handler in self._event_handlers[evnt]:
                if event_handler == evnt_handler:
                    return True
        return False

    def remove_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Removes an event handler of a given event.

        Note that if the same event handler was added multiple times
        the event, all the duplicated handlers are removed. This
        method relies on the ``__eq__`` method of the input event
        handler to compare event handlers.

        Args:
            event (str): Specifies the event handler is attached to.
            event_handler (``BaseEventHandler``): Specifies the event
                handler to remove.

        Raises:
            ValueError: if the event does not exist or if the handler
                is not attached to the event.

        Example usage:

        .. code-block:: python

            # Create an event manager
            >>> from gravitorch.utils.events import EventManager
            >>> event_manager = EventManager()
            # Add an event handler to the engine
            >>> def hello_handler():
            ...     print('Hello!')
            ...
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> event_manager.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            True
            # Remove the event handler of the engine
            >>> event_manager.remove_event_handler('my_event', VanillaEventHandler(hello_handler))
            # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            False
        """
        if event not in self._event_handlers:
            raise ValueError(f"'{event}' event does not exist")

        new_event_handlers = [
            handler for handler in self._event_handlers[event] if event_handler != handler
        ]
        if len(new_event_handlers) == len(self._event_handlers[event]):
            raise ValueError(
                f"{event_handler} is not found among registered event handlers for '{event}' event"
            )
        if len(new_event_handlers) > 0:
            self._event_handlers[event] = new_event_handlers
        else:
            del self._event_handlers[event]
        logger.debug(f"Removed {event_handler} in '{event}' event")

    def reset(self) -> None:
        r"""Resets the event manager.

        This method removes all the event handlers from the event manager.

        Example usage:

        .. code-block:: python

            # Create an event manager
            >>> from gravitorch.utils.events import EventManager
            >>> event_manager = EventManager()
            # Add an event handler to the engine
            >>> def hello_handler():
            ...     print('Hello!')
            ...
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> event_manager.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            True
            >>> event_manager.fire_event('my_event')
            >>> event_manager.last_fired_event
            my_event
            # Reset the event manager
            >>> event_manager.reset()
            # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> event_manager.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            False
            >>> event_manager.last_fired_event
            None
        """
        self._event_handlers.clear()
        self._last_fired_event = None


def to_event_handlers_str(event_handlers: dict[str, list], num_spaces: int = 2):
    r"""Computes a string representation of the event handlers.

    Args:
        event_handlers (dict): Specifies the dictionary with the list
            of handlers for each event.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
        str: A string representation of the event handlers
    """
    lines = []
    spaces = " " * num_spaces
    for key, value in event_handlers.items():
        lines.append(
            f"({key})\n{spaces}"
            f"{str_add_indent(to_torch_sequence_str(value, num_spaces), num_spaces)}"
        )
    return "\n".join(lines)



r"""This module defines some utility functions to format some objects."""

__all__ = [
    "convert_human_readable_count",
    "convert_seconds_to_readable_format",
    "str_add_indent",
    "str_scalar",
    "str_target_object",
    "to_flat_dict",
    "to_human_readable_byte_size",
    "to_pretty_dict_str",
    "to_pretty_json_str",
    "to_pretty_yaml_str",
    "to_torch_mapping_str",
    "to_torch_sequence_str",
]

import datetime
import json
import math
from collections.abc import Mapping, Sequence
from typing import Any, Optional, TypeVar, Union

import yaml
from objectory import OBJECT_TARGET

PARAMETER_NUM_UNITS = (" ", "K", "M", "B", "T")

BYTE_UNITS = {
    "B": 1,
    "KB": 1024,
    "MB": 1024 * 1024,
    "GB": 1024 * 1024 * 1024,
    "TB": 1024 * 1024 * 1024 * 1024,
}

T = TypeVar("T")


def convert_seconds_to_readable_format(seconds: Union[int, float]) -> str:
    r"""Converts a number of seconds in an easier format to read hh:mm:ss.

    If the number of seconds is bigger than 1 day, this representation
    also encodes the number of days.

    Args:
        seconds (integer or float): Specifies the number of seconds.

    Returns:
        str: The number of seconds in a string format (hh:mm:ss).

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import convert_seconds_to_readable_format
        >>> convert_seconds_to_readable_format(1.2)
        '0:00:01.200000'
        >>> convert_seconds_to_readable_format(61.2)
        '0:01:01.200000'
        >>> convert_seconds_to_readable_format(3661.2)
        '1:01:01.200000'
    """
    return str(datetime.timedelta(seconds=seconds))


def str_add_indent(original: Any, num_spaces: int = 2) -> str:
    r"""Add indentations if the original string is a multi-lines string.

    Args:
        original: Specifies the original string. If the inputis not a
            string, it will be converted to a string with the function
            ``str``.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
        str: The indented string.

    Example usage:

    .. code-block:: python

        >>> string_to_print = "string1\nstring2"
        >>> print(f"{string_to_print}")
        string1
        string2
        >>> print(f"\t{string_to_print}")
            string1
        string2
        # The problem is that 'string1' and 'string2' are not aligned.
        # The indentation is only applied to the first line
        >>> from gravitorch.utils.format import str_add_indent
        >>> print(f"\t{str_add_indent(string_to_print, 4)}")
            string1
            string2
    """
    formatted_str = str(original).split("\n")
    if len(formatted_str) == 1:  # single line
        return formatted_str[0]
    first = formatted_str.pop(0)
    formatted_str = "\n".join([(num_spaces * " ") + line for line in formatted_str])
    return first + "\n" + formatted_str


def str_scalar(value: Union[int, float]) -> str:
    r"""Returns a string representation of a scalar value.

    Args:
        value (int or float): Specifies the input value.

    Returns:
        str: The string representation of the input value.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import str_scalar
        >>> str_scalar(123456.789)
        123,456.789000
        >>> str_scalar(1234567)
        1,234,567
        >>> str_scalar(12345678901)
        1.234568e+10
    """
    if isinstance(value, int):
        if math.fabs(value) >= 1e9:
            return f"{value:.6e}"
        return f"{value:,}"
    if math.fabs(value) < 1e-3 or math.fabs(value) >= 1e6:
        return f"{value:.6e}"
    return f"{value:,.6f}"


def str_target_object(config: dict) -> str:
    r"""Gets a string that indicates the target object in the config.

    Args:
        config (dict): Specifies a config using the ``object_factory``
            library. This dict is expected to have a key
            ``'_target_'`` to indicate the target object.

    Returns:
        str: A string with the target object.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import str_target_object
        >>> str_target_object({OBJECT_TARGET: "something.MyClass"})
        [_target_: something.MyClass]
        >>> str_target_object({})
        [_target_: N/A]
    """
    return f"[{OBJECT_TARGET}: {config.get(OBJECT_TARGET, 'N/A')}]"


def convert_human_readable_count(number: Union[int, float]) -> str:
    r"""Converts an integer number with K, M, B, T for thousands, millions,
    billions and trillions, respectively.

    Args:
        number (int or float): A positive integer number. If the
            number is a float, it will be converted to an integer.

    Returns:
        str: A string formatted according to the pattern described
            above.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import convert_human_readable_count
        >>> convert_human_readable_count(123)
        '123  '
        >>> convert_human_readable_count(1234)  # (one thousand)
        '1.2 K'
        >>> convert_human_readable_count(2e6)   # (two million)
        '2.0 M'
        >>> convert_human_readable_count(3e9)   # (three billion)
        '3.0 B'
        >>> convert_human_readable_count(4e14)  # (four hundred trillion)
        '400 T'
        >>> convert_human_readable_count(5e15)  # (more than trillion)
        '5,000 T'
    """
    if number < 0:
        raise ValueError(f"The number should be a positive number (received {number})")
    labels = PARAMETER_NUM_UNITS
    num_digits = int(math.floor(math.log10(number)) + 1 if number > 0 else 1)
    num_groups = min(
        int(math.ceil(num_digits / 3)), len(labels)
    )  # don't abbreviate beyond trillions
    shift = -3 * (num_groups - 1)
    number = number * (10**shift)
    index = num_groups - 1
    if index < 1 or number >= 100:
        return f"{int(number):,d} {labels[index]}"
    return f"{number:,.1f} {labels[index]}"


def to_flat_dict(
    data: Union[dict, list, tuple, str, bool, int, float, None],
    prefix: Optional[str] = None,
    separator: str = ".",
    to_str: Union[type[object], tuple[type[object], ...], None] = None,
) -> dict[str, Union[str, bool, int, float, None]]:
    r"""Computes a flat representation of a nested dict with the dot format.

    Args:
        data: Specifies the nested dict to flat.
        prefix (str, optional): Specifies the prefix to use to
            generate the name of the key. ``None`` means no prefix.
            Default: ``None``
        separator (str, optional): Specifies the separator to
            concatenate keys of nested collections. Default: ``'.'``
        to_str (tuple or ``None``, optional): Specifies the data types
            which will not be flattened out, instead converted to
            string. Default: ``None``

    Returns:
        dict: The flatted data.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_flat_dict
        >>> data = {
                'str': 'def',
                'module': {
                    'component': {
                        'float': 3.5,
                        'int': 2,
                    },
                },
            }
        >>> to_flat_dict(data)
        {
            'module.component.float': 3.5,
            'module.component.int': 2,
            'str': 'def',
        }
        # Example with lists (also works with tuple)
        >>> data = {
                'module': [[1, 2, 3], {'bool': True}],
                'str': 'abc',
            }
        >>> to_flat_dict(data)
        {
            'module.0.0': 1,
            'module.0.1': 2,
            'module.0.2': 3,
            'module.1.bool': True,
            'str': 'abc',
        }

        # Example with lists with to_str=(list) (also works with tuple)
        >>> data = {
                'module': [[1, 2, 3], {'bool': True}],
                'str': 'abc',
            }
        >>> to_flat_dict(data)
        {
            'module': "[[1, 2, 3], {'bool': True}]",
            'str': 'abc',
        }
    """
    flat_dict = {}
    to_str = to_str or tuple()
    if isinstance(data, to_str):
        flat_dict[prefix] = str(data)
    elif isinstance(data, dict):
        for key, value in data.items():
            flat_dict.update(
                to_flat_dict(
                    value,
                    prefix=f"{prefix}{separator}{key}" if prefix else key,
                    separator=separator,
                    to_str=to_str,
                )
            )
    elif isinstance(data, (list, tuple)):
        for i, value in enumerate(data):
            flat_dict.update(
                to_flat_dict(
                    value,
                    prefix=f"{prefix}{separator}{i}" if prefix else str(i),
                    separator=separator,
                    to_str=to_str,
                )
            )
    else:
        flat_dict[prefix] = data
    return flat_dict


def to_pretty_json_str(
    data: Any, sort_keys: bool = True, indent: int = 2, max_len: int = 80
) -> str:
    r"""Converts a data structure to a pretty JSON string.

    Args:
        data: Specifies the input to convert to a pretty JSON string.
        sort_keys (bool, optional): Specifies if the keys are sorted
            or not. Default: ``True``
        indent (int, optional): Specifies the indent. It is a
            non-negative integer. Default: ``2``
        max_len (int, optional): Specifies the maximum length of the
            string representation. If the string representation is
            longer than this length, it is converted to the json
            representation. Default: ``80``

    Returns:
        str: The string representation.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_pretty_json_str
        >>> to_pretty_json_str({"my_key": "my_value"})
        "{'my_key': 'my_value'}"
        >>> to_pretty_json_str(["value1", "value2"])
        "['value1', 'value2']"
        >>> to_pretty_json_str(["value1", "value2"], max_len=5)
        '[\n  "value1",\n  "value2"\n]'
    """
    str_data = str(data)
    if len(str_data) < max_len:
        return str_data
    return json.dumps(data, sort_keys=sort_keys, indent=indent, default=str)


def to_pretty_yaml_str(
    data: Any, sort_keys: bool = True, indent: int = 2, max_len: int = 80
) -> str:
    r"""Converts a data structure to a pretty YAML string.

    Args:
        data: Specifies the input to convert to a pretty YAML string.
        sort_keys (bool, optional): Specifies if the keys are sorted
            or not. Default: ``True``
        indent (int, optional): Specifies the indent. It is a
            non-negative integer. Default: ``2``
        max_len (int, optional): Specifies the maximum length of the
            string representation. If the string representation is
            longer than this length, it is converted to the json
            representation. Default: ``max_len``

    Returns:
        str: The string representation.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_pretty_yaml_str
        >>> to_pretty_yaml_str({"my_key": "my_value"})
        "{'my_key': 'my_value'}"
        >>> to_pretty_yaml_str(["value1", "value2"])
        "['value1', 'value2']"
        >>> to_pretty_yaml_str(["value1", "value2"], max_len=5)
        '- value1\n- value2\n'
    """
    str_data = str(data)
    if len(str_data) < max_len:
        return str_data
    return yaml.safe_dump(data, sort_keys=sort_keys, indent=indent)


def to_pretty_dict_str(data: dict[str, Any], sorted_keys: bool = False, indent: int = 0) -> str:
    r"""Converts a dict to a pretty string representation.

    This function was designed for flat dictionary. If you have a
    nested dictionary, you may consider other functions. Note that
    this function works for nested dict but the output may not be
    nice.

    Args:
        data (dict): Specifies the input dictionary.
        sorted_keys (bool, optional): Specifies if the key of the dict
            are sorted or not. Default: ``False``
        indent (int, optional): Specifies the indentation. The value
            should be greater or equal to 0. Default: ``0``

    Returns:
        str: The string representation.

        Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_pretty_dict_str
        >>> to_pretty_dict_str({"my_key": "my_value"})
        'my_key : my_value'
        >>> to_pretty_dict_str({"key1": "value1", "key2": "value2"})
        'key1 : value1\nkey2 : value2'
    """
    if indent < 0:
        raise ValueError(f"The indent has to be greater or equal to 0 (received: {indent})")
    if not data:
        return ""

    max_length = max([len(key) for key in data.keys()])
    output = []
    for key in sorted(data.keys()) if sorted_keys else data.keys():
        output.append(f"{' ' * indent + str(key) + ' ' * (max_length - len(key))} : {data[key]}")
    return "\n".join(output)


def to_torch_mapping_str(mapping: Mapping, sorted_keys: bool = False, num_spaces: int = 2) -> str:
    r"""Computes a torch-like (``torch.nn.Module``) string representation of a
    mapping.

    Args:
        mapping (``Mapping``): Specifies the mapping.
        sorted_keys (bool, optional): Specifies if the key of the dict
            are sorted or not. Default: ``False``
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
        str: The string representation of the mapping.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_torch_mapping_str
        >>> to_torch_mapping_str({'key1': 'abc', 'key2': 'something\nelse'})
        (key1) abc
        (key2) something
          else
    """
    lines = []
    for key, value in sorted(mapping.items()) if sorted_keys else mapping.items():
        lines.append(f"({key}) {str_add_indent(value, num_spaces=num_spaces)}")
    return "\n".join(lines)


def to_torch_sequence_str(sequence: Sequence, num_spaces: int = 2) -> str:
    r"""Computes a torch-like (``torch.nn.Module``) string representation of a
    sequence.

    Args:
        sequence (``Sequence``): Specifies the sequence.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
        str: The string representation of the sequence.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_torch_sequence_str
        >>> to_torch_sequence_str(["abc", "something\nelse"])
        (0) abc
        (1) something
          else
    """
    lines = []
    for i, item in enumerate(sequence):
        lines.append(f"({i}) {str_add_indent(item, num_spaces=num_spaces)}")
    return "\n".join(lines)


def to_human_readable_byte_size(size: int, unit: Optional[str] = None) -> str:
    r"""Gets a human-readable representation of the byte size.

    Args:
        size (int): Specifies the size in bytes.
        unit (str, optional): Specifies the unit. If ``None``, the
            best unit is found automatically. The supported units
            are: ``'B'``, ``'KB'``, ``'MB'``, ``'GB'``, ``'TB'``.
            Default: ``None``

    Returns:
        str: The byte size in a human-readable format.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.format import to_human_readable_byte_size
        >>> to_human_readable_byte_size(2)
        2.00 B
        >>> to_human_readable_byte_size(2048)
        2.00 KB
        >>> to_human_readable_byte_size(2097152)
        2.00 MB
        >>> to_human_readable_byte_size(2048, unit='B')
        2,048.00 B
    """
    if unit is None:  # Find the best unit.
        best_unit = "B"
        for unit, multiplier in BYTE_UNITS.items():
            if (size / multiplier) > 1:
                best_unit = unit
        unit = best_unit

    if unit not in BYTE_UNITS:
        raise ValueError(
            f"Incorrect unit {unit}. The available units are {list(BYTE_UNITS.keys())}"
        )

    return f"{size / BYTE_UNITS.get(unit, 1):,.2f} {unit}"



r"""This package contains the implementation of some datasets."""

__all__ = [
    "DemoMultiClassClsDataset",
    "FileToInMemoryDataset",
    "ImageFolderDataset",
    "InMemoryDataset",
    "MNISTDataset",
    "log_box_dataset_class",
    "setup_dataset",
]

from gravitorch.data.datasets.demo_map_style import DemoMultiClassClsDataset
from gravitorch.data.datasets.factory import setup_dataset
from gravitorch.data.datasets.image_folder import ImageFolderDataset
from gravitorch.data.datasets.in_memory import FileToInMemoryDataset, InMemoryDataset
from gravitorch.data.datasets.mnist import MNISTDataset
from gravitorch.data.datasets.utils import log_box_dataset_class
